# Copyright (C) 2006-2008 Red Hat, Inc.
#
# This program is free software; you can redistribute
# it and/or modify it under the terms of version 2 of the
# GNU General Public License as published by the
# Free Software Foundation.

from xml.dom import minidom, Node
from TagObject import TagObject
from Cluster import Cluster
from ClusterNode import ClusterNode
from ClusterNodes import ClusterNodes
from Fence import Fence
from FenceDevice import FenceDevice
from FenceDevices import FenceDevices
from Method import Method
from Device import Device
from Cman import Cman
from Totem import Totem
from Gulm import Gulm
from Lockserver import Lockserver
from Ip import Ip
from Script import Script
from NFSClient import NFSClient
from NFSExport import NFSExport
from Fs import Fs
from Samba import Samba
from Apache import Apache
from LVM import LVM
from MySQL import MySQL
from OpenLDAP import OpenLDAP
from Postgres8 import Postgres8
from Tomcat5 import Tomcat5
from SAPDatabase import SAPDatabase
from SAPInstance import SAPInstance
from SybaseASE import SybaseASE 
from OracleDB import OracleDB
from Multicast import Multicast
from FenceDaemon import FenceDaemon
from FenceXVMd import FenceXVMd
from Netfs import Netfs
from Clusterfs import Clusterfs
from Resources import Resources
from Service import Service
from QuorumD import QuorumD
from Heuristic import Heuristic
from Vm import Vm
from RefObject import RefObject
from FailoverDomain import FailoverDomain
from FailoverDomains import FailoverDomains
from FailoverDomainNode import FailoverDomainNode
from Rm import Rm
from Altname import Altname

DLM_TYPE    = 0
GULM_TYPE   = 1

TAGNAMES = { 'cluster': Cluster,
             'clusternodes': ClusterNodes,
             'clusternode': ClusterNode,
             'altname': Altname,
             'fence': Fence,
             'fencedevice': FenceDevice,
             'fencedevices': FenceDevices,
             'method': Method,
             'cman': Cman,
             'totem': Totem,
             'gulm': Gulm,
             'lockserver': Lockserver,
             'rm': Rm,
             'service': Service,
             'vm': Vm,
             'fence_xvmd': FenceXVMd,
             'resources': Resources,
             'failoverdomain': FailoverDomain,
             'failoverdomains': FailoverDomains,
             'failoverdomainnode': FailoverDomainNode,
             'ip': Ip,
             'fs': Fs,
             'smb': Samba,
             'apache': Apache,
             'lvm': LVM,
             'mysql': MySQL,
             'openldap': OpenLDAP,
             'postgres-8': Postgres8,
             'tomcat-5': Tomcat5,
             'fence_daemon': FenceDaemon,
             'multicast': Multicast,
             'clusterfs': Clusterfs,
             'netfs': Netfs,
             'quorumd': QuorumD,
             'heuristic': Heuristic,
             'script': Script,
             'nfsexport': NFSExport,
             'nfsclient': NFSClient,
             'SAPDatabase': SAPDatabase,
             'SAPInstance': SAPInstance,
             'ASEHAagent': SybaseASE,
             'oracledb': OracleDB,
             'device': Device }


###---Don't translate strings below---
CLUSTER_PTR_STR = "cluster"
CLUSTERNODES_PTR_STR = "clusternodes"
FAILDOMS_PTR_STR = "failoverdomains"
FENCEDEVICES_PTR_STR = "fencedevices"
RESOURCEMANAGER_PTR_STR = "rm"
RESOURCES_PTR_STR = "resources"
FENCEDAEMON_PTR_STR = "fence_daemon"
SERVICE = "service"
VM = "vm"
FENCE_XVMD_STR = "fence_xvmd"
GULM_TAG_STR = "gulm"
MCAST_STR = "multicast"
CMAN_PTR_STR = "cman"
TOTEM_PTR_STR = "totem"
QUORUMD_PTR_STR = "quorumd"
###-----------------------------------


INVALID_GULM_COUNT = "GULM locking mechanism may consist of 1, 3, or 5 locking servers. You have configured %d. Fix the error and try saving again."


class ModelBuilder:
  def __init__(self, mcast_addr=None, domm=None, cluster_os=None):
    if domm is None:
      raise Exception, 'No cluster configuration'

    self.lock_type = DLM_TYPE
    self.mcast_address = mcast_addr
    self.mcast_interface = None
    self.mcast_ptr = None
    self.cluster_ptr = None
    self.GULM_ptr = None
    self.CMAN_ptr = None
    self.TOTEM_ptr = None
    self.clusternodes_ptr = None
    self.failoverdomains_ptr = None
    self.fencedevices_ptr = None
    self.resourcemanager_ptr = None
    self.resources_ptr = None
    self.fence_daemon_ptr = None
    self.isModified = False
    self.quorumd_ptr = None
    self.usesQuorumd = False
    self.fence_xvmd_ptr = None
    self.unusual_items = list()
    self.isVirtualized = False
    self.cluster_os = cluster_os 

    if mcast_addr is None:
      self.usesMulticast = False
    else:
      self.usesMulticast = True

    self.parent = domm
    self.object_tree = self.buildModel(None)
    self.check_empty_ptrs()
    self.check_fence_daemon()
    self.resolve_fence_instance_types()
    self.purgePCDuplicates()
    self.resolve_references()
    self.check_for_multicast()
    self.check_for_nodeids()

  def getClusterOS(self):
    return self.cluster_os

  def buildModel(self, parent_node, parent_object=None):

    if parent_node is None:
      parent_node = self.parent

    new_object = None
    if parent_node.nodeType == Node.DOCUMENT_NODE:
      parent_node = parent_node.firstChild

    if parent_node.nodeType == Node.ELEMENT_NODE:
      #Create proper type
      try:
        new_object = TAGNAMES[parent_node.nodeName]()
        attrs = parent_node.attributes
        for attrName in attrs.keys():
          attrNode = attrs.get(attrName)
          attrValue = attrNode.nodeValue
          new_object.addAttribute(attrName, attrValue)
      except KeyError, e: ##This allows for custom tags
        new_object = TagObject(parent_node.nodeName)
        attrs = parent_node.attributes
        for attrName in attrs.keys():
          attrNode = attrs.get(attrName)
          attrValue = attrNode.nodeValue
          new_object.addAttribute(attrName, attrValue)
        self.unusual_items.append((parent_object, new_object))
        for item in parent_node.childNodes:
          result_object = self.buildModel(item, new_object)
          if result_object is not None:
            new_object.addChild(result_object)
        return None

       ######End of unusual item exception

      if parent_node.nodeName == CLUSTER_PTR_STR:
        self.cluster_ptr = new_object
      if parent_node.nodeName == CLUSTERNODES_PTR_STR:
        self.clusternodes_ptr = new_object
      elif parent_node.nodeName == FENCEDEVICES_PTR_STR:
        self.fencedevices_ptr = new_object
      elif parent_node.nodeName == FAILDOMS_PTR_STR:
        self.failoverdomains_ptr = new_object
      elif parent_node.nodeName == RESOURCEMANAGER_PTR_STR:
        self.resourcemanager_ptr = new_object
      elif parent_node.nodeName == RESOURCES_PTR_STR:
        self.resources_ptr = new_object
      elif parent_node.nodeName == FENCEDAEMON_PTR_STR:
        self.fence_daemon_ptr = new_object
      elif parent_node.nodeName == QUORUMD_PTR_STR:
        self.quorumd_ptr = new_object
        self.usesQuorumd = True
      elif parent_node.nodeName == GULM_TAG_STR:
        self.GULM_ptr = new_object
        self.lock_type = GULM_TYPE
      elif parent_node.nodeName == CMAN_PTR_STR:
        self.CMAN_ptr = new_object
      elif parent_node.nodeName == TOTEM_PTR_STR:
        self.TOTEM_ptr = new_object
      elif parent_node.nodeName == MCAST_STR:
        self.usesMulticast = True
      elif parent_node.nodeName == FENCE_XVMD_STR:
        self.fence_xvmd_ptr = new_object

    else:
      return None


    for item in parent_node.childNodes:
      result_object = self.buildModel(item, new_object)
      if result_object is not None:
        new_object.addChild(result_object)

    return (new_object)

  def buildDLMModelTemplate(self):
    obj_tree = Cluster()
    self.cluster_ptr = obj_tree

    obj_tree.addAttribute("name", "alpha_cluster")
    obj_tree.addAttribute("config_version", "1")
    fdp = FenceDaemon()
    obj_tree.addChild(fdp)
    self.fence_daemon_ptr = fdp
    cns = ClusterNodes()
    obj_tree.addChild(cns)
    self.clusternodes_ptr = cns

    cman = Cman()
    self.CMAN_ptr = cman
    obj_tree.addChild(cman)

    if self.usesMulticast is True:
      mcast = Multicast()
      mcast.addAttribute("addr", self.mcast_address)
      if self.mcast_interface is not None:
        mcast.addAttribute("interface", self.mcast_interface)
      cman.addChild(mcast)
      self.mcast_ptr = mcast

    fds = FenceDevices()
    obj_tree.addChild(fds)
    self.fencedevices_ptr = fds

    rm = Rm()
    obj_tree.addChild(rm)
    self.resourcemanager_ptr = rm

    fdoms = FailoverDomains()
    self.failoverdomains_ptr = fdoms
    rm.addChild(fdoms)

    rcs = Resources()
    rm.addChild(rcs)
    self.resources_ptr = rcs

    self.isModified = False

    return obj_tree

  def buildGULMModelTemplate(self):
    obj_tree = Cluster()
    self.cluster_ptr = obj_tree

    obj_tree.addAttribute("name", "alpha_cluster")
    obj_tree.addAttribute("config_version", "1")
    cns = ClusterNodes()
    obj_tree.addChild(cns)
    self.clusternodes_ptr = cns

    gulm = Gulm()
    self.GULM_ptr = gulm
    self.fence_daemon_ptr = None
    obj_tree.addChild(gulm)

    fds = FenceDevices()
    obj_tree.addChild(fds)
    self.fencedevices_ptr = fds

    rm = Rm()
    obj_tree.addChild(rm)
    self.resourcemanager_ptr = rm

    fdoms = FailoverDomains()
    self.failoverdomains_ptr = fdoms
    rm.addChild(fdoms)

    rcs = Resources()
    rm.addChild(rcs)
    self.resources_ptr = rcs

    self.isModified = False

    return obj_tree

  def getFenceDeviceByName(self, name):
    device = filter(lambda x: x.getName() == name, self.getFenceDevices())
    num_fdevs = len(device)
    if num_fdevs > 1:
      raise Exception, '%d fence devices named %s exist' % (num_fdevs, name)
    if num_fdevs < 1:
      return None
    return device[0]
      
  ##Because fence devices are declared in a separate XML section
  ##in conf file, agent types for fence instances must be done in
  ##a separate pass, after the DOM is completely built. This method
  ##sets the agent type for each fence instance.
  def resolve_fence_instance_types(self):
    agent_hash = {}
    for fd in self.getFenceDevices():
      agent = fd.getAgentType()
      if agent is not None:
        agent_hash[fd.getName()] = agent

    for node in self.getNodes():
      for level in node.getFenceLevels():
        for child in level.getChildren():
          child.setAgentType(agent_hash[child.getName()])
    
  ##This method builds RefObject containers for appropriate
  ##entities after the object tree is built.
  def resolve_references(self):
    reset_list_sentinel = True
    while(reset_list_sentinel is True):
      reset_list_sentinel = False
      resource_children = self.resourcemanager_ptr.getChildren()
      for r_child in resource_children:
        if r_child.getTagName() == SERVICE:
         reset_list_sentinel = self.find_references(r_child)
         if reset_list_sentinel is True:
           break

  def find_references(self, entity, parent=None):
    result = False
    if (entity.getAttribute("ref") is not None) and (entity.isRefObject() is False):
      result = self.transform_reference(entity, parent)
      return result

    children = entity.getChildren()
    if len(children) > 0:
      for child in children:
        result = self.find_references(child, entity)
        if result is True:
          return result

    return result

  def transform_reference(self, entity, parent):
    result = False
    #This entity has a "ref" attr...need to walk through resources list
    #and look for a match
    recs = self.resources_ptr.getChildren()
    for rec in recs:
      if rec.getTagName() == entity.getTagName():
        if entity.getTagName() == "ip":
          if entity.getAttribute("ref") == rec.getAttribute("address"):
            rf = RefObject(rec)
            kids = entity.getChildren()
            for kid in kids:
              rf.addChild(kid)
            result = True
            break
        else:
          if entity.getAttribute("ref") == rec.getName():
            rf = RefObject(rec)
            kids = entity.getChildren()
            for kid in kids:
              rf.addChild(kid)
            result = True
            break

    if result is False:
      return result

    try:
      entity_attr = entity.getAttributes()
      if entity_attr is not None:
        for i in entity_attr.iterkeys():
          if not rf.attr_hash.has_key(i):
            rf.addAttribute(i, entity_attr[i])
    except:
      pass

    if parent is None:  #Must be a service
      self.resourcemanager_ptr.addChild(rf)
      self.resourcemanager_ptr.removeChild(entity)
      return True
    else:
      parent.addChild(rf)
      parent.removeChild(entity)
      return True

  def exportModelAsString(self):
    if self.perform_final_check() is False: # failed
      return None

    #check for dual power fences
    self.dual_power_fence_check()
    self.restore_unusual_items()

    try:
      doc = minidom.Document()
      self.object_tree.generateXML(doc)
      strbuf = doc.toprettyxml()

      self.isModified = False

      ##Need to restore model
      self.parent = doc

      self.object_tree = self.buildModel(None)
      self.check_empty_ptrs()
      self.check_fence_daemon()
      self.resolve_fence_instance_types()
      self.purgePCDuplicates()
      self.resolve_references()
      self.check_for_multicast()

    finally:
      pass
      #dual_power_fence_check() adds extra
      #fence instance entries for dual power controllers
      #These must be removed from the tree before the UI
      #can be used
      #self.purgePCDuplicates()

    return strbuf

  def restore_unusual_items(self):
    for item in self.unusual_items:
      duplicate = False
      kids = item[0].getChildren()
      for kid in kids:
        if kid == item[1]:
          duplicate = True
          break
      if duplicate is True:
        continue
      else:
        item[0].addChild(item[1])

  def check_for_nodeids(self):
    nodes = self.getNodes()
    for node in nodes:
      if node.getAttribute('nodeid') is None:
        new_id = self.getUniqueNodeID()
        node.addAttribute('nodeid', new_id)

  def getUniqueNodeID(self):
    nodes = self.getNodes()
    total_nodes = len(nodes)
    dex_list = list()
    for nd_idx in range (1, (total_nodes + 3)):
      dex_list.append(str(nd_idx))

    for dex in dex_list:
      found = False
      for node in nodes:
        ndid = node.getAttribute('nodeid')
        if ndid is not None:
          if ndid == dex:
            found = True
            break
        else:
          continue

      if found is True:
        continue
      else:
        return dex

  def setIsVirtualized(self, isVirtualized):
    if isVirtualized is None:
      self.isVirtualized = False
    else:
      self.isVirtualized = isVirtualized

  def getIsVirtualized(self):
    return self.isVirtualized

  def getNodes(self):
    #Find the clusternodes obj and return get_children
    return self.clusternodes_ptr.getChildren()

  def getNodeNames(self):
    return map(lambda x: x.getName(), self.clusternodes_ptr.getChildren())

  def getNodeNameById(self, node_id):
    return filter(lambda x: x.getAttribute('nodeid') == node_id, self.clusternodes_ptr.getChildren())[0].getName()

  def addNode(self, clusternode):
    self.clusternodes_ptr.addChild(clusternode)
    if self.usesMulticast is True:
      mcast = Multicast()
      mcast.addAttribute('addr', self.mcast_address)
      if self.mcast_interface is not None:
        mcast.addAttribute('interface', self.mcast_interface)
      clusternode.addChild(mcast)
    self.isModified = True

  def deleteNode(self, clusternode):
    #1) delete any non-shared fence devices used by this node
    #2) delete node
    #3) delete failoverdomainnodes with same name
    #4) delete lockserver nodes if GULM

    name = clusternode.getName()

    for level in clusternode.getFenceLevels():
        for fence in level.getChildren():
            fdev = self.getFenceDeviceByName(fence.getName())
            if fdev and not fdev.isShared():
                self.deleteFenceDevice(fdev)
                
    self.clusternodes_ptr.removeChild(clusternode)

    found_one = True

    while found_one is True:
      found_one = False
      fdoms = self.getFailoverDomains()
      for fdom in fdoms:
        children = fdom.getChildren()
        for child in children:
          if child.getName() == name:
            fdom.removeChild(child)
            found_one = True
            break

      lock_type = self.getLockType()
      if lock_type == GULM_TYPE:
        if self.isNodeLockserver(clusternode.getName()) is True:
          self.removeLockserver(clusternode)

    self.isModified = True

  def retrieveNodeByName(self, name):
    ret = filter(lambda x: x.getName() == name, self.clusternodes_ptr.getChildren())
    if len(ret) != 1:
      raise KeyError, name
    return ret[0]

  def deleteNodeByName(self, name):
    return self.deleteNode(self.retrieveNodeByName(name))

  def retrieveServiceByName(self, name):
    svcs = self.getServices()
    for svc in svcs:
      if svc.getName() == name:
        return svc

    raise KeyError, 'Couldn\'t find service name %s in current list' % name

  def retrieveVMsByName(self, name):
    vms = self.getVMs()
    for v in vms:
      if v.getName() == name:
        return v

    raise KeyError, 'Couldn\'t find VM name %s in current list' % name

  def del_totem(self):
    if self.TOTEM_ptr is not None:
      self.cluster_ptr.removeChild(self.TOTEM_ptr)
      self.TOTEM_ptr = None

  def add_totem(self, obj):
    self.del_totem()
    if self.TOTEM_ptr is None:
      self.cluster_ptr.addChild(obj)
      self.TOTEM_ptr = obj

  def hasFenceXVM(self):
    return self.fence_xvmd_ptr is not None

  # Right now the fence_xvmd tag is empty, but allow the object
  # to be passed in case attributes are added in the future.
  def addFenceXVM(self, obj):
    if self.fence_xvmd_ptr is not None:
      self.cluster_ptr.removeChild(self.fence_xvmd_ptr)
    self.cluster_ptr.addChild(obj)
    self.fence_xvmd_ptr = obj

  def delFenceXVM(self):
    if self.fence_xvmd_ptr is not None:
      self.cluster_ptr.removeChild(self.fence_xvmd_ptr)
      self.fence_xvmd_ptr = None

  def addFenceDevice(self, device):
    self.fencedevices_ptr.addChild(device)

  def deleteFenceDevice(self, device):
    if self.fencedevices_ptr is None or device is None:
      return False
    self.fencedevices_ptr.removeChild(device)
    return True

  def getFenceDevices(self):
    if self.fencedevices_ptr is None:
      return list()
    else:
      return self.fencedevices_ptr.getChildren()

  def getFenceDevicePtr(self):
    return self.fencedevices_ptr

  def getFailoverDomains(self):
    if self.failoverdomains_ptr is None:
      return list()
    else:
      return self.failoverdomains_ptr.getChildren()

  def getFailoverDomainPtr(self):
    return self.failoverdomains_ptr

  def getFailoverDomainByName(self, fdom_name):
    fdoms = self.getFailoverDomains()
    for i in fdoms:
      if i.getName() == fdom_name:
        return i
    return None

  def getFailoverDomainsForNode(self, nodename):
    matches = list()
    faildoms = self.getFailoverDomains()
    for fdom in faildoms:
      fdnodes = fdom.getChildren()
      for node in fdnodes:
        if node.getName().encode('ascii', 'ignore') == nodename:
          matches.append(fdom)
          break

    return matches

  def addNodeToFailoverDomain(self, node, fd):
    fdoms = self.getFailoverDomains()
    for fdom in fdoms:
      if fdom.getName().encode('ascii', 'ignore') == fd:
        kids = fdom.getChildren()
        newnode = FailoverDomainNode()
        if len(kids) != 0: #Use an existing node as a baseline...
          attrs = kids[0].getAttributes()
          kees = attrs.keys()
          for k in kees:
            newnode.addAttribute(k, attrs[k])
            newnode.addAttribute("name", node)
        else:
          newnode.addAttribute("name", node)
        fdom.addChild(newnode)
        return

    return

  def removeNodeFromFailoverDomain(self, node, fd):
    fdoms = self.getFailoverDomains()
    for fdom in fdoms:
      if fdom.getName().encode('ascii', 'ignore') == fd:
        kids = fdom.getChildren()
        for kid in kids:
          if kid.getName().encode('ascii', 'ignore') == node:
            fdom.removeChild(kid)
            return

    return

  def isMulticast(self):
    return self.usesMulticast

  def check_for_multicast(self):
    if self.usesMulticast is True:
      #set mcast address
      children = self.CMAN_ptr.getChildren()
      for child in children:
        if child.getTagName() == MCAST_STR:
          self.mcast_ptr = child
          addr = child.getAttribute("addr")
          if addr is not None:
            self.mcast_address = addr
          else:  #What a mess! a multicast tag, but no addr attribute
            self.mcast_address = ""
          mcastif = child.getAttribute("interface")
          if mcastif is not None:
            self.mcast_interface = mcastif
          return

  def getMcastAddr(self):
    return self.mcast_address

  def getMcastInterface(self):
    return self.mcast_interface

  def isQuorumd(self):
    return self.usesQuorumd

  def getQuorumdPtr(self):
    return self.quorumd_ptr

  def check_empty_ptrs(self):
    if self.resourcemanager_ptr is None:
      rm = Rm()
      self.cluster_ptr.addChild(rm)
      self.resourcemanager_ptr = rm

    if self.failoverdomains_ptr is None:
      fdoms = FailoverDomains()
      self.resourcemanager_ptr.addChild(fdoms)
      self.failoverdomains_ptr = fdoms

    if self.fencedevices_ptr is None:
      fds = FenceDevices()
      self.cluster_ptr.addChild(fds)
      self.fencedevices_ptr = fds

    if self.resources_ptr is None:
      rcs = Resources()
      self.resourcemanager_ptr.addChild(rcs)
      self.resources_ptr = rcs

    if self.GULM_ptr is None and self.fence_daemon_ptr is None:
      fdp = FenceDaemon()
      self.cluster_ptr.addChild(fdp)
      self.fence_daemon_ptr = fdp

  def getServices(self):
    rg_list = list()
    if self.resourcemanager_ptr is not None:
      kids = self.resourcemanager_ptr.getChildren()
      for kid in kids:
        if kid.getTagName() == SERVICE:
          rg_list.append(kid)

    return rg_list

  def getService(self, svcname):
    services = self.getServices()
    for service in services:
      if service.getName() == svcname:
        return service

    return None

  def getVMs(self):
    rg_list = list()
    if self.resourcemanager_ptr is not None:
      kids = self.resourcemanager_ptr.getChildren()
      for kid in kids:
        if kid.getTagName() == VM:
          rg_list.append(kid)

    return rg_list

  def deleteService(self, service):
    if self.resourcemanager_ptr is not None:
      kids = self.resourcemanager_ptr.getChildren()
      for kid in kids:
        if kid.getName() == service:
          self.resourcemanager_ptr.removeChild(kid)
          break

  def getResources(self):
    if self.resources_ptr is not None:
      return self.resources_ptr.getChildren()
    else:
      return list()

  def getResourcesPtr(self):
    return self.resources_ptr

  def getResourceManagerPtr(self):
    return self.resourcemanager_ptr

  def getResourceByName(self, name):
    resources = self.resources_ptr.getChildren()
    res = filter(lambda x: x.getName() == name, resources)
    if not res or len(res) < 1:
      raise KeyError, name
    if len(res) > 1:
      raise Exception, 'More than one resource is named "%s"' % name
    return res[0]

  def deleteResource(self, name):
    for i in self.resources_ptr.getChildren():
      if i.getName() == name:
        self.resources_ptr.removeChild(i)
        return i
    raise KeyError, name

  def getClusterNodesPtr(self):
    return self.clusternodes_ptr

  def getClusterPtr(self):
    return self.cluster_ptr

  def getClusterName(self):
    return self.getClusterPtr().getName()

  def getClusterAlias(self):
    return self.getClusterPtr().getAlias()

  def getGULMPtr(self):
    return self.GULM_ptr

  def getCMANPtr(self):
    return self.CMAN_ptr

  def addTotemPtr(self):
    cp = self.getClusterPtr()
    totem = Totem()
    cp.addChild(totem)
    return totem

  def getTotemPtr(self):
    return self.TOTEM_ptr

  def getLockServer(self, name):
    children = self.GULM_ptr.getChildren()
    for child in children:
      if child.getName() == name:
        return child

    return None

  def getLockType(self):
    return self.lock_type

  def isNodeLockserver(self, name):
    gptr = self.getGULMPtr()
    if gptr is None:  #Obviously not GULM
      return False
    children = gptr.getChildren()
    for child in children:
      if child.getName() == name:
        return True

    return False

  def removeLockserver(self, clusternode):
    gptr = self.getGULMPtr()
    if gptr is None:  #Obviously not GULM
      return
    children = gptr.getChildren()
    for child in children:
      if child.getName() == clusternode.getName():
        gptr.removeChild(child)
        break  #Only one will be found

    self.isModified = True

  def switch_lockservers(self):
    #first get what type of locking is currently in place
    if self.lock_type == DLM_TYPE:
      #remove <cman>
      self.cluster_ptr.removeChild(self.CMAN_ptr)
      if self.TOTEM_ptr is not None:
        self.del_totem()
      self.CMAN_ptr = None

      #add gulm tag
      gulm = Gulm()
      self.GULM_ptr = gulm
      self.cluster_ptr.addChild(gulm)

      #check for multicast
      #if multicast, remove <multicast> from each node
      #remove votes attr from each node
      nodes = self.getNodes()
      for node in nodes:
        if self.usesMulticast is True:
          mnode = node.getMulticastNode()
          if mnode is not None:
            node.removeChild(mnode)
        node.removeAttribute('votes')

      self.usesMulticast = None
      self.mcast_address = None
      self.mcast_interface = None
      self.mcast_ptr = None

      #reset self.lock_type
      self.lock_type = GULM_TYPE

      #make the first node a lockserver
      nodes = self.getNodes()
      for node in nodes:
        ls = Lockserver()
        ls.addAttribute('name', node.getName())
        self.GULM_ptr.addChild(ls)
        break


      #set modified
      self.isModified = True

    else:
      #remove <gulm> tag
      children = self.cluster_ptr.getChildren()
      for child in children:
        if child.getTagName() == "gulm":
          self.cluster_ptr.removeChild(child)
          break

      #set gulm pointer to None
      self.GULM_ptr = None

      #add <cman> tag
      cman = Cman()
      self.CMAN_ptr = cman
      self.cluster_ptr.addChild(cman)

      #reset self.lock_type
      self.lock_type = DLM_TYPE

      #give each node a vote of 1
      nds = self.getNodes()
      for nd in nds:
        nd.addAttribute('votes', '1')

      #set modified
      self.isModified = True

  def set_nodes_multicast(self, mcast_addr, mcast_if=None):
    for node in iter(self.getNodes()):
      new_mcast_tag = True
      mcast = None

      for node_child in iter(node.getChildren()):
        if node_child.getTagName() == MCAST_STR:
          mcast = node_child
          new_mcast_tag = False
          break

      if mcast is None:
        mcast = Multicast()

      mcast.addAttribute('addr', mcast_addr)
      if mcast_if:
        mcast.addAttribute('interface', mcast_if)
      else:
        try:
          mcast.removeAttribute('interface')
        except:
          pass

      if new_mcast_tag is True:
        node.addChild(mcast)

  def set_cluster_multicast(self, mcast_addr, mcast_if=None):
    if not self.CMAN_ptr:
      return False

    if self.mcast_ptr is None:
      mcast = Multicast()
      self.CMAN_ptr.addChild(mcast)
      self.mcast_ptr = mcast
    else:
      mcast = self.mcast_ptr

    mcast.addAttribute('addr', mcast_addr)
    self.mcast_address = mcast_addr
    self.mcast_interface = mcast_if
    self.usesMulticast = True
    self.set_nodes_multicast(mcast_addr, mcast_if=mcast_if)
    self.isModified = True

  def del_nodes_multicast(self):
    for node in iter(self.getNodes()):
      for node_child in iter(node.getChildren()):
        if node_child.getTagName() == MCAST_STR:
          node.removeChild(node_child)
          break

  def del_cluster_multicast(self):
    if self.CMAN_ptr is None:
      return False

    for child in iter(self.CMAN_ptr.getChildren()):
      if child.getTagName() == MCAST_STR:
        self.CMAN_ptr.removeChild(child)

    self.mcast_ptr = None
    self.usesMulticast = False
    self.mcast_address = None
    self.mcast_interface = None
    self.del_nodes_multicast()
    self.isModified = True

  def swap_multicast_state(self, address=None, mcast_if=None):
    if self.usesMulticast is True:
      self.del_cluster_multicast()
    else:
      self.set_cluster_multicast(address, mcast_if)

  def check_fence_daemon(self):
    if self.GULM_ptr is None and self.fence_daemon_ptr is None:
      self.fence_daemon_ptr = FenceDaemon()
      self.cluster_ptr.addChild(self.fence_daemon_ptr)

  def getFenceDaemonPtr(self):
    return self.fence_daemon_ptr

  def isFileModified(self):
    return self.isModified

  def setModified(self, modified=None):
    if modified is None:
      self.isModified = True
    else:
      self.isModified = modified

  def rectifyNewNodenameWithFaildoms(self, oldname, newname):
    fdoms = self.getFailoverDomains()
    for fdom in fdoms:
      children = fdom.getChildren() #These are failoverdomainnodes...
      for child in children:
        if child.getName().strip() == oldname:
          child.addAttribute("name", newname)

  ###This method runs through ALL fences (Device objs) and changes name attr
  ###to new name
  def rectifyNewFencedevicenameWithFences(self, oldname, newname):
    nodes = self.getNodes()
    for node in nodes:
      levels = node.getFenceLevels()
      for level in levels:
        fences = level.getChildren()
        for fence in fences:
          if fence.getName() == oldname:
            fence.addAttribute("name", newname)

  ###Method for removing fence instances if a fence device
  ###has been deleted from the configuration
  def removeFenceInstancesForFenceDevice(self, name):
    nodes = self.getNodes()
    for node in nodes:
      levels = node.getFenceLevels()
      for level in levels:
        fences = level.getChildren()
        kill_list = list()
        for fence in fences:
          if fence.getName() == name:
            kill_list.append(fence)
        for victim in kill_list:
          level.removeChild(victim)

  def removeReferences(self, tagobj):
    self.__removeReferences(tagobj, self.cluster_ptr)
  def __removeReferences(self, tagobj, level):
    for t in level.getChildren()[:]:
      if t.isRefObject():
        if t.getObj() == tagobj:
          level.removeChild(t)
          continue
      self.__removeReferences(tagobj, t)

  def updateReferences(self):
    self.__updateReferences(self.cluster_ptr)
  def __updateReferences(self, level):
    for t in level.getChildren():
      if t.isRefObject():
        t.setRef(t.getObj().getName())
      self.__updateReferences(t)

  def perform_final_check(self):
    if self.check_gulm_count() is False:
      return False
    self.check_two_node()

    #add more checks

    return True

  def check_gulm_count(self):
    if self.getLockType() == GULM_TYPE:
      gulm_count = len(self.getGULMPtr().getChildren())
      if not (gulm_count in (1, 3, 5)):
        return False
    return True

  def check_two_node(self):
    if self.getLockType() == DLM_TYPE and self.quorumd_ptr is None:
      clusternodes_count = len(self.clusternodes_ptr.getChildren())
      #Make certain that there is a cman tag in the file
      #If missing, it will not hurt to add it here
      if self.CMAN_ptr is None:
        cman = Cman()
        self.cluster_ptr.addChild(cman)
        self.CMAN_ptr = cman
      if clusternodes_count == 2:
        self.CMAN_ptr.addAttribute('two_node', '1')
        self.CMAN_ptr.addAttribute('expected_votes', '1')
      else:
        if self.CMAN_ptr.getAttribute('expected_votes') in ('0', '1'):
          self.CMAN_ptr.removeAttribute('expected_votes')
        self.CMAN_ptr.removeAttribute('two_node')

  def dual_power_fence_check(self):
    # if 2 or more power controllers reside in the same fence level,
    # duplicate entries must be made for every controller with an
    # attribute for option set first for off, then for on.

    # for every node:
      # for every fence level:
        # examine every fence
        # If fence is of power type, add to 'found' list for that level
        # If 'found' list is longer than 1, write out extra objs
    for node in self.getNodes():
      for level in node.getFenceLevels():
        pc_list = list()
        for kid in level.getChildren():
          if kid.hasNativeOptionSet() is True:
            continue
          if kid.isPowerController() is True:
            pc_list.append(kid)

        if len(pc_list) > 1:
          # Means we found multiple PCs in the same level
          for fence in pc_list:
            fence.addAttribute("option", "off")
            d = Device()
            d.setAgentType(fence.getAgentType())
            attrs = fence.getAttributes()
            for (k, v) in attrs.iteritems():
              d.addAttribute(k, v)
            d.addAttribute("option", "on")
            level.addChild(d)
        
  def purgePCDuplicates(self):
    found_one = True
    while found_one is True:
      found_one = False
      nodes = self.getNodes()
      for node in nodes:
        levels = node.getFenceLevels()
        for level in levels:
          kids = level.getChildren()
          for kid in kids: #kids are actual fence instance objects
            #Need to pass over this device if:
            ##1) It is not a power controller, or
            ##2) It had an initial option attr when the model was constructed
            if not kid.isPowerController():
              continue
            if kid.hasNativeOptionSet():
              continue
            res = kid.getAttribute("option")
            if res is not None:
              if res == "off":
                kid.removeAttribute("option")
              else:
                level.removeChild(kid)
                found_one = True
                break
        if found_one is True:
          break

  def searchObjectTree(self, tagtype):
    objlist = list()
    self.object_tree.searchTree(objlist, tagtype)

    return objlist
