# Copyright (C) 2006-2008 Red Hat, Inc.
#
# This program is free software; you can redistribute
# it and/or modify it under the terms of version 2 of the
# GNU General Public License as published by the
# Free Software Foundation.

TAG_NAME = "document"

class TagObject:
  def __init__(self, tagname=None):
    self.attr_hash = {}
    self.children = list()
    if tagname is None:
      self.TAG_NAME = TAG_NAME
    else:
      self.TAG_NAME = tagname

  def addChild(self, child):
    self.children.append(child)

  def removeChild(self, child):
    self.children.remove(child)

  def addAttribute(self, name, value):
    self.attr_hash[name] = value

  def removeAttribute(self, key):
    try:
      del(self.attr_hash[key])
    except KeyError, e:
      return False

    return True

  def generateXML(self, doc, parent=None):
    tag = doc.createElement(self.TAG_NAME)
    if parent is None:
      parent = doc
    parent.appendChild(tag)
    self.exportAttributes(tag)
    if len(self.children) > 0:
      for child in self.children:
        if child is not None:
          child.generateXML(doc, tag)

  def exportAttributes(self, tag):
    attrs = self.attr_hash.keys()
    for attr in attrs:
      tag.setAttribute(attr, self.attr_hash[attr])

  def getAttributes(self):
    return self.attr_hash

  def getAttribute(self, kee):
    try:
      return self.attr_hash[kee]
    except KeyError, e:
      return None

  def getChildren(self):
    return self.children

  def getName(self):
    try:
      return self.attr_hash["name"]
    except KeyError, e:
      return ""

  def getTagName(self):
    return self.TAG_NAME

  def isRefObject(self):
    return False

  def searchTree(self, objlist, tagtype):
    if self.TAG_NAME == tagtype:
      objlist.append(self)
    if len(self.children) > 0:
      for child in self.children:
        if child is None:
          continue
        child.searchTree(objlist, tagtype)
