# Copyright (C) 2006-2008 Red Hat, Inc.
#
# This program is free software; you can redistribute
# it and/or modify it under the terms of version 2 of the
# GNU General Public License as published by the
# Free Software Foundation.

from Products.Archetypes.utils import make_uuid
from ClusterModel.ModelBuilder import ModelBuilder
import RicciQueries as rq
from ricci_communicator import RicciCommunicator
from LuciSyslog import get_logger
from LuciDB import resolve_nodename
from LuciZope import GetReqVars

from conga_constants import CLUSTER_CONFIG, CLUSTER_DELETE, \
	CLUSTER_PROCESS, CLUSTER_RESTART, CLUSTER_START, CLUSTER_STOP, \
	NODE_FORCE_DELETE, FDOM, FENCEDEV, NODE, NODE_ACTIVE, \
	NODE_ACTIVE_STR, NODE_DELETE, NODE_FENCE, NODE_INACTIVE, \
	NODE_INACTIVE_STR, NODE_JOIN_CLUSTER, NODE_LEAVE_CLUSTER, \
	NODE_PROCESS, NODE_REBOOT, NODE_UNKNOWN, NODE_UNKNOWN_STR, \
	PROP_FENCE_TAB, PROP_GENERAL_TAB, PROP_GULM_TAB, PROP_MCAST_TAB, \
	PROP_QDISK_TAB, RESOURCE, RESOURCE_CONFIG, RESOURCE_REMOVE, \
	SERVICE, SERVICE_DELETE, SERVICE_MIGRATE, SERVICE_RESTART, \
	SERVICE_START, SERVICE_STOP, VM_CONFIG, \
	LUCI_DEBUG_MODE, LUCI_CLUSTER_BASE_URL

luci_log = get_logger()

def getnodes(self, model):
	try:
		return map(lambda x: str(x.getName()), model.getNodes())
	except Exception, e:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('getnodes0: %r %s' % (e, str(e)))
	return []

def getResourceInfo(model, name, baseurl, res=None):
	if res is None:
		try:
			res = model.getResourceByName(name)
		except Exception, e:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('GRI0: %s: %r %s' % (name, e, str(e)))
			return {}

	res_info = {}
	res_name = res.getName().strip()

	res_info['name'] = res_name
	res_info['attrs'] = res.attr_hash
	res_info['type'] = res.resource_type
	res_info['tag_name'] = res.TAG_NAME

	cluname = model.getClusterName()
	res_info['cfgurl'] = '%s?clustername=%s&resourcename=%s&pagetype=%s' \
		% (baseurl, cluname, res_name, RESOURCE_CONFIG)
	res_info['url'] = '%s?clustername=%s&resourcename=%s&pagetype=%s' \
		% (baseurl, cluname, res_name, RESOURCE)
	res_info['delurl'] = '%s?clustername=%s&resourcename=%s&pagetype=%s' \
		% (baseurl, cluname, res_name, RESOURCE_REMOVE)
	return res_info

def getResources(model, baseurl):
	if model is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('getResources0: model is none')
		return []
	return map(lambda x: getResourceInfo(model, None, baseurl, x), model.getResources())

def getClusterStatusModel(model):
	results = list()
	vals = {}

	try:
		clustername = model.getClusterName()
		clusteralias = model.getClusterAlias()
		vals['type'] = 'cluster'
		vals['alias'] = clusteralias
		vals['name'] = clustername
		vals['error'] = True
		vals['votes'] = '[unknown]'
		vals['quorate'] = '[unknown]'
		vals['minQuorum'] = '[unknown]'
		results.append(vals)
	except Exception, e:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('GCSM0: %r %s' % (e, str(e)))
		return None

	try:
		nodelist = model.getNodes()
	except Exception, e:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('GCSM1: %r %s' % (e, str(e)))
		return None

	for node in nodelist:
		node_val = {}
		node_val['type'] = 'node'
		try:
			node_name = node.getName()
			if not node_name:
				raise Exception, 'cluster node name is unknown'
		except:
			node_name = '[unknown]'

		node_val['name'] = node_name
		node_val['clustered'] = '[unknown]'
		node_val['online'] = '[unknown]'
		node_val['error'] = True

		try:
			votes = node.getVotes()
			if not votes:
				raise Exception, 'unknown unmber of votes'
		except:
			votes = '[unknown]'

		node_val['votes'] = votes
		results.append(node_val)
	return results

def getClusterStatus(self, request, rc, cluname=None):
	try:
		doc = rq.getClusterStatusBatch(rc) or None
	except Exception, e:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('GCS0: error: %r: %r %s' \
				% (cluname, e, str(e)))
		doc = None

	if doc is None:
		model = LuciExtractCluModel(self, request, cluname)
		if model is not None:
			try:
				cinfo = getClusterStatusModel(model)
				if not cinfo or len(cinfo) < 1:
					raise Exception, 'cinfo is None'
				return cinfo
			except Exception, e:
				if LUCI_DEBUG_MODE is True:
					luci_log.debug_verbose('GCS1: %r: %r %s' \
						% (cluname, e, str(e)))

	if doc is None:
		try:
			from LuciDB import getClusterStatusDB
			if cluname:
				clustername = cluname
			else:
				fvars = GetReqVars(request, [ 'clustername' ])

				clustername = fvars['clustername']
				if clustername is None:
					raise Exception, 'unable to determine cluster name'

			cinfo = getClusterStatusDB(self, clustername)
			if not cinfo or len(cinfo) < 1:
				raise Exception, 'cinfo is None'
			return cinfo
		except Exception, e:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('GCS2: cluster %r info from DB: %r %s' \
					% (cluname, e, str(e)))
		return []

	results = list()
	vals = {}
	vals['type'] = 'cluster'

	try:
		vals['alias'] = doc.firstChild.getAttribute('alias')
	except AttributeError, e:
		vals['alias'] = doc.firstChild.getAttribute('name')

	vals['votes'] = doc.firstChild.getAttribute('votes')
	vals['name'] = doc.firstChild.getAttribute('name')
	vals['minQuorum'] = doc.firstChild.getAttribute('minQuorum')
	vals['quorate'] = doc.firstChild.getAttribute('quorate')
	results.append(vals)

	for node in doc.firstChild.childNodes:
		if node.nodeName == 'node':
			vals = {}
			vals['type'] = 'node'
			vals['clustered'] = node.getAttribute('clustered')
			vals['name'] = node.getAttribute('name')
			vals['online'] = node.getAttribute('online')
			vals['uptime'] = node.getAttribute('uptime')
			vals['votes'] = node.getAttribute('votes')
			results.append(vals)
		elif node.nodeName == 'service':
			vals = {}
			vals['type'] = 'service'
			vals['name'] = node.getAttribute('name')
			vals['nodename'] = node.getAttribute('nodename')
			vals['running'] = node.getAttribute('running')
			try:
				vals['is_vm'] = node.getAttribute('vm').lower() == 'true'
			except:
				vals['is_vm'] = False
			vals['failed'] = node.getAttribute('failed')
			vals['autostart'] = node.getAttribute('autostart')
			results.append(vals)
	return results

def getServicesInfo(self, status, model, req):
	svc_map = {}
	maplist = list()
	fvars = GetReqVars(req, [ 'clustername', 'URL' ])

	baseurl = fvars['URL'] or LUCI_CLUSTER_BASE_URL

	cluname = fvars['clustername']
	if cluname is None:
		cluname = model.getClusterName()

	nodes = model.getNodes()
	for item in status:
		if item['type'] == 'service':
			itemmap = {}
			itemmap['name'] = item['name']

			cur_node = None
			if item['running'] == 'true':
				cur_node = item['nodename']
				itemmap['running'] = 'true'
				itemmap['nodename'] = cur_node
				itemmap['disableurl'] = '%s?clustername=%s&servicename=%s&pagetype=%s' % (baseurl, cluname, item['name'], SERVICE_STOP)
				itemmap['restarturl'] = '%s?clustername=%s&servicename=%s&pagetype=%s' % (baseurl, cluname, item['name'], SERVICE_RESTART)
			else:
				itemmap['enableurl'] = '%s?clustername=%s&servicename=%s&pagetype=%s' % (baseurl, cluname, item['name'], SERVICE_START)

			itemmap['autostart'] = item['autostart']

			try:
				svc = model.retrieveServiceByName(item['name'])
				itemmap['cfgurl'] = '%s?clustername=%s&servicename=%s&pagetype=%s' % (baseurl, cluname, item['name'], SERVICE)
				itemmap['delurl'] = '%s?clustername=%s&servicename=%s&pagetype=%s' % (baseurl, cluname, item['name'], SERVICE_DELETE)
			except:
				try:
					svc = model.retrieveVMsByName(item['name'])
					itemmap['is_vm'] = True
					itemmap['cfgurl'] = '%s?clustername=%s&servicename=%s&pagetype=%s' % (baseurl, cluname, item['name'], VM_CONFIG)
					itemmap['delurl'] = '%s?clustername=%s&servicename=%s&pagetype=%s' % (baseurl, cluname, item['name'], VM_CONFIG)
				except:
					continue

			starturls = list()
			for node in nodes:
				cur_nodename = node.getName()
				if node.getName() != cur_node:
					starturl = {}
					starturl['nodename'] = cur_nodename
					starturl['url'] = '%s?clustername=%s&servicename=%s&pagetype=%s&nodename=%s' % (baseurl, cluname, item['name'], SERVICE_START, cur_nodename)
					starturls.append(starturl)

					if itemmap.has_key('is_vm') and itemmap['is_vm'] is True:
						migrate_url = { 'nodename': cur_nodename }
						migrate_url['migrate'] = True
						migrate_url['url'] = '%s?clustername=%s&servicename=%s&pagetype=%s&nodename=%s' % (baseurl, cluname, item['name'], SERVICE_MIGRATE, cur_nodename)
						starturls.append(migrate_url)

			itemmap['links'] = starturls

			dom = svc.getAttribute('domain')
			if dom is not None:
				itemmap['faildom'] = dom
			else:
				itemmap['faildom'] = 'No Failover Domain'
			maplist.append(itemmap)

	svc_map['services'] = maplist
	return svc_map

def recurse_resources(parent_uuid, child, resource_list, indent_ctr, parent=None):
	#First, add the incoming child as a resource
	#Next, check for children of it
	#Call yourself on every children
	#then return

	rc_map = {}
	if parent is not None:
		rc_map['parent'] = parent
	rc_map['name'] = child.getName()

	#Note: Final version needs all resource attrs
	if child.isRefObject() is True:
		child_obj = child.getObj()
		rc_map['attrs'] = child.getAttributes().copy()
		rc_map['attrs'].update(child_obj.getAttributes())
		rc_map['type'] = child_obj.getResourceType()
		rc_map['ref_object'] = True
		rc_map['tag_name'] = child_obj.TAG_NAME
	else:
		rc_map['tag_name'] = child.TAG_NAME
		rc_map['type'] = child.getResourceType()
		rc_map['attrs'] = child.getAttributes()

	rc_map['indent_ctr'] = indent_ctr

	rc_map['uuid'] = make_uuid('resource')
	rc_map['parent_uuid'] = parent_uuid

	resource_list.append(rc_map)
	kids = child.getChildren()
	child_depth = 0
	new_indent_ctr = indent_ctr + 1
	for kid in kids:
		cdepth = recurse_resources(rc_map['uuid'], kid, resource_list, new_indent_ctr, child)
		child_depth = max(cdepth, child_depth)

	rc_map['max_depth'] = child_depth
	return child_depth + 1


def get_fdom_names(model):
	return map(lambda x: x.getName(), model.getFailoverDomains())


def getServiceInfo(self, status, model, req):
	root_uuid = 'toplevel'

	fvars = GetReqVars(req, [ 'clustername', 'servicename', 'URL' ])

	baseurl = fvars['URL'] or LUCI_CLUSTER_BASE_URL
	if not model:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('getServiceInfo0: no model: %r' % model)
		return {}

	#set up struct for service config page
	hmap = {}

	try:
		cluname = fvars['clustername'] or model.getClusterName()
		hmap['fdoms'] = get_fdom_names(model)
	except Exception, e:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('getServiceInfo1: %r %s' % (e, str(e)))
		hmap['fdoms'] = list()

	hmap['root_uuid'] = root_uuid
	# uuids for the service page needed when new resources are created
	hmap['uuid_list'] = map(lambda x: make_uuid('resource'), xrange(30))

	servicename = fvars['servicename']
	if servicename is None:
		return hmap

	if len(status) > 0:
		nodenames = model.getNodeNames()

	for item in status:
		innermap = {}
		if item['type'] == 'service':
			if item['name'] == servicename:
				hmap['name'] = servicename
				hmap['autostart'] = item['autostart']

				starturls = list()
				if item['running'] == 'true':
					hmap['running'] = 'true'
					nodename = item['nodename']
					innermap['current'] = 'Running on %s' % nodename

					innermap['disableurl'] = '%s?clustername=%s&servicename=%s&pagetype=%s' % (baseurl, cluname, servicename, SERVICE_STOP)
					innermap['restarturl'] = '%s?clustername=%s&servicename=%s&pagetype=%s' % (baseurl, cluname, servicename, SERVICE_RESTART)
					innermap['delurl'] = '%s?clustername=%s&servicename=%s&pagetype=%s' % (baseurl, cluname, servicename, SERVICE_DELETE)

					# In this case, determine where it can run...
					for node in nodenames:
						if node != nodename:
							starturl = {}
							starturl['nodename'] = node
							starturl['url'] = '%s?clustername=%s&servicename=%s&pagetype=%s&nodename=%s' % (baseurl, cluname, servicename, SERVICE_START, node)
							starturls.append(starturl)

							if item.has_key('is_vm') and item['is_vm'] is True:
								migrate_url = { 'nodename': node }
								migrate_url['url'] = '%s?clustername=%s&servicename=%s&pagetype=%s&nodename=%s' % (baseurl, cluname, servicename, SERVICE_MIGRATE, node)
								migrate_url['migrate'] = True
								starturls.append(migrate_url)
					innermap['links'] = starturls
				else:
					#Do not set ['running'] in this case...ZPT will detect it is missing
					innermap['current'] = 'Stopped'
					innermap['enableurl'] = '%s?clustername=%s&servicename=%s&pagetype=%s' % (baseurl, cluname, servicename, SERVICE_START)
					innermap['delurl'] = '%s?clustername=%s&servicename=%s&pagetype=%s' % (baseurl, cluname, servicename, SERVICE_DELETE)

					starturls = list()
					for node in nodenames:
						starturl = {}

						starturl['nodename'] = node
						starturl['url'] = '%s?clustername=%s&servicename=%s&pagetype=%s&nodename=%s' % (baseurl, cluname, servicename, SERVICE_START, node)
						starturls.append(starturl)

						if item.has_key('is_vm') and item['is_vm'] is True:
							migrate_url = { 'nodename': node }
							migrate_url['url'] = '%s?clustername=%s&servicename=%s&pagetype=%s&nodename=%s' % (baseurl, cluname, servicename, SERVICE_MIGRATE, node)
							migrate_url['migrate'] = True
							starturls.append(migrate_url)
					innermap['links'] = starturls
				hmap['innermap'] = innermap

	# Now build hashes for resources under service.
	# first get service by name from model

	svc = model.getService(servicename)
	try:
		hmap['domain'] = svc.getAttribute('domain')
	except:
		hmap['domain'] = None

	try:
		hmap['recovery'] = svc.getAttribute('recovery')
	except:
		hmap['recovery'] = None

	try:
		if int(svc.getAttribute('exclusive')):
			hmap['exclusive'] = 'true'
		else:
			hmap['exclusive'] = 'false'
	except:
		hmap['exclusive'] = 'false'

	resource_list = list()
	if svc is not None:
		indent_ctr = 0
		children = svc.getChildren()
		for child in children:
			recurse_resources(root_uuid, child, resource_list, indent_ctr)

	hmap['resource_list'] = resource_list
	return hmap

def getFdomInfo(self, model, request):
	fhash = {}
	fhash['members'] = {}

	try:
		fdom = model.getFailoverDomainByName(request['fdomname'])
		fhash['name'] = fdom.getName()
	except Exception, e:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('getFdomInfo0: %r %s' % (e, str(e)))
		return fhash

	ordered_attr = fdom.getAttribute('ordered')
	if ordered_attr is not None and (ordered_attr == 'true' or ordered_attr == '1'):
		fhash['prioritized'] = '1'
	else:
		fhash['prioritized'] = '0'

	
	nofailback = fdom.getAttribute('nofailback')
	if nofailback is not None and (nofailback == 'true' or nofailback == '1'):
		fhash['nofailback'] = '1'
	else:
		fhash['nofailback'] = '0'

	restricted_attr = fdom.getAttribute('restricted')
	if restricted_attr is not None and (restricted_attr == 'true' or restricted_attr == '1'):
		fhash['restricted'] = '1'
	else:
		fhash['restricted'] = '0'

	nodes = fdom.getChildren()
	for node in nodes:
		try:
			priority = node.getAttribute('priority')
		except:
			priority = '1'
		fhash['members'][node.getName()] = { 'priority': priority }
	return fhash

def getFdomsInfo(self, model, request, clustatus):
	fvars = GetReqVars(request, [ 'clustername', 'URL' ])

	baseurl = fvars['URL'] or LUCI_CLUSTER_BASE_URL
	clustername = fvars['clustername']
	if clustername is None:
		return {}

	slist = list()
	nlist = list()
	for item in clustatus:
		if item['type'] == 'node':
			nlist.append(item)
		elif item['type'] == 'service':
			slist.append(item)

	fdomlist = list()
	for fdom in model.getFailoverDomains():
		fdom_map = {}
		fdom_name = fdom.getName()
		fdom_map['name'] = fdom_name
		fdom_map['cfgurl'] = '%s?pagetype=%s&clustername=%s&fdomname=%s' \
			% (baseurl, FDOM, clustername, fdom.getName())

		ordered_attr = fdom.getAttribute('ordered')
		if ordered_attr is not None and (ordered_attr == 'true' or ordered_attr == '1'):
			fdom_map['ordered'] = True
		else:
			fdom_map['ordered'] = False

		restricted_attr = fdom.getAttribute('restricted')
		if restricted_attr is not None and (restricted_attr == 'true' or restricted_attr == '1'):
			fdom_map['restricted'] = True
		else:
			fdom_map['restricted'] = False

		nofailback = fdom.getAttribute('nofailback')
		if nofailback is not None and (nofailback == 'true' or nofailback == '1'):
			fdom_map['nofailback'] = True
		else:
			fdom_map['nofailback'] = False

		nodelist = list()
		for node in fdom.getChildren():
			nodesmap = {}
			ndname = node.getName()

			for nitem in nlist:
				if nitem['name'] == ndname:
					nodesmap['nodename'] = ndname
					nodesmap['nodecfgurl'] = '%s?clustername=%s&nodename=%s&pagetype=%s' \
						% (baseurl, clustername, ndname, NODE)
					if nitem['clustered'] == 'true':
						nodesmap['status'] = NODE_ACTIVE
					elif nitem['online'] == 'false':
						nodesmap['status'] = NODE_UNKNOWN
					else:
						nodesmap['status'] = NODE_INACTIVE
					priority_attr =	node.getAttribute('priority')
					if priority_attr is not None:
						nodesmap['priority'] = '0'
					nodelist.append(nodesmap)
		fdom_map['nodeslist'] = nodelist

		svclist = list()
		tmp = model.getServices()
		tmp.extend(model.getVMs())
		for svc in tmp:
			svcname = svc.getName()
			for sitem in slist:
				if sitem['name'] == svcname:
					domain = svc.getAttribute('domain')
					if domain == fdom_name:
						svcmap = {}
						svcmap['name'] = svcname
						svcmap['status'] = sitem['running']
						if svc.getTagName() == 'vm':
							svcmap['svcurl'] = '%s?pagetype=%s&clustername=%s&servicename=%s' % (baseurl, VM_CONFIG, clustername, svcname)
						else:
							svcmap['svcurl'] = '%s?pagetype=%s&clustername=%s&servicename=%s' % (baseurl, SERVICE, clustername, svcname)
						svcmap['location'] = sitem['nodename']
						svclist.append(svcmap)
		fdom_map['svclist'] = svclist
		fdomlist.append(fdom_map)

	return fdomlist

def getClusterInfo(self, model, req):
	fvars = GetReqVars(req, [ 'clustername', 'URL' ])

	baseurl = fvars['URL'] or LUCI_CLUSTER_BASE_URL
	cluname = fvars['clustername']
	if cluname is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('GCI0: unable to determine cluster name')
		return {}

	clumap = {}
	try:
		cluster_os = model.getClusterOS()
		if cluster_os.find('Tikanga'):
			os_minor = int(cluster_os[cluster_os.find('5.') + 2])
			if os_minor > 1:
				clumap['vm_migration_choice'] = True
	except:
		pass

	if not model:
		try:
			model = getModelForCluster(self, cluname)
			if not model:
				raise Exception, 'model is none'
			req.SESSION.set('model', model)
		except Exception, e:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('GCI1: unable to get model for cluster %s: %r %s' % (cluname, e, str(e)))
			return {}
	else:
		totem = model.getTotemPtr()
		if totem:
			clumap['totem'] = totem.getAttributes()

	prop_baseurl = '%s?pagetype=%s&clustername=%s&' \
		% (baseurl, CLUSTER_CONFIG, cluname)
	basecluster_url = '%stab=%s' % (prop_baseurl, PROP_GENERAL_TAB)
	# needed:
	clumap['basecluster_url'] = basecluster_url
	# name field
	clumap['clustername'] = model.getClusterAlias()
	# config version
	cp = model.getClusterPtr()
	clumap['config_version'] = cp.getConfigVersion()

	# xvmd info
	clumap['fence_xvmd'] = model.hasFenceXVM()

	#-------------
	#new cluster params - if rhel5
	#-------------
	gulm_ptr = model.getGULMPtr()
	if not gulm_ptr:
		#Fence Daemon Props
		fencedaemon_url = '%stab=%s' % (prop_baseurl, PROP_FENCE_TAB)
		clumap['fencedaemon_url'] = fencedaemon_url
		fdp = model.getFenceDaemonPtr()
		pjd = fdp.getAttribute('post_join_delay')
		if pjd is None:
			pjd = '6'
		pfd = fdp.getAttribute('post_fail_delay')
		if pfd is None:
			pfd = '0'
		#post join delay
		clumap['pjd'] = pjd
		#post fail delay
		clumap['pfd'] = pfd

		#-------------
		#if multicast
		multicast_url = '%stab=%s' % (prop_baseurl, PROP_MCAST_TAB)
		clumap['multicast_url'] = multicast_url
		#mcast addr
		is_mcast = model.isMulticast()
		if is_mcast:
			clumap['mcast_addr'] = model.getMcastAddr()
			clumap['mcast_interface'] = model.getMcastInterface()
			clumap['is_mcast'] = 'True'
		else:
			clumap['mcast_addr'] = None
			clumap['mcast_interface'] = None
			clumap['is_mcast'] = 'False'
		clumap['gulm'] = False
	else:
		#-------------
		# GULM params (RHEL4 only)
		#-------------
		lockserv_list = list()
		clunodes = model.getNodes()
		gulm_lockservs = map(lambda x: x.getName(), gulm_ptr.getChildren())
		lockserv_list = map(lambda x: (x, True), gulm_lockservs)
		for node in clunodes:
			n = node.getName()
			if not n in gulm_lockservs:
				lockserv_list.append((n, False))
		clumap['gulm'] = True
		clumap['gulm_url'] = '%stab=%s' % (prop_baseurl, PROP_GULM_TAB)
		clumap['gulm_lockservers'] = lockserv_list

	#-------------
	# quorum disk params
	#-------------
	quorumd_url = '%stab=%s' % (prop_baseurl, PROP_QDISK_TAB)
	clumap['quorumd_url'] = quorumd_url
	is_quorumd = model.isQuorumd()
	clumap['is_quorumd'] = is_quorumd
	clumap['interval'] = ''
	clumap['tko'] = ''
	clumap['votes'] = ''
	clumap['min_score'] = ''
	clumap['device'] = ''
	clumap['label'] = ''

	# list struct for heuristics
	hlist = list()
	if is_quorumd:
		qdp = model.getQuorumdPtr()
		interval = qdp.getAttribute('interval')
		if interval is not None:
			clumap['interval'] = interval

		tko = qdp.getAttribute('tko')
		if tko is not None:
			clumap['tko'] = tko

		votes = qdp.getAttribute('votes')
		if votes is not None:
			clumap['votes'] = votes

		min_score = qdp.getAttribute('min_score')
		if min_score is not None:
			clumap['min_score'] = min_score

		device = qdp.getAttribute('device')
		if device is not None:
			clumap['device'] = device

		label = qdp.getAttribute('label')
		if label is not None:
			clumap['label'] = label

		heuristic_kids = qdp.getChildren()
		for kid in heuristic_kids:
			hmap = {}
			hprog = kid.getAttribute('program')
			if hprog is None:
				continue

			hscore = kid.getAttribute('score')
			hmap['hprog'] = hprog
			if hscore is not None:
				hmap['hscore'] = hscore
			else:
				hmap['hscore'] = ''

			hinterval = kid.getAttribute('interval')
			if hinterval is not None:
				hmap['hinterval'] = hinterval
			else:
				hmap['hinterval'] = ''
			hlist.append(hmap)
	clumap['hlist'] = hlist

	return clumap

def getClustersInfo(self, status, req):
	clu_map = {}

	fvars = GetReqVars(req, [ 'URL' ])
	baseurl = fvars['URL'] or LUCI_CLUSTER_BASE_URL

	nodelist = list()
	svclist = list()
	clulist = list()
	for item in status:
		if item['type'] == 'node':
			nodelist.append(item)
		elif item['type'] == 'service':
			svclist.append(item)
		elif item['type'] == 'cluster':
			clulist.append(item)
		else:
			continue

	if len(clulist) < 1:
		return {}
	clu = clulist[0]
	if clu.has_key('error'):
		clu_map['error'] = clu['error']
	if clu.has_key('errmsg'):
		clu_map['errmsg'] = clu['errmsg']
	clustername = clu['name']
	if not clu['alias']:
		clu_map['clusteralias'] = clu['alias']
	else:
		clu_map['clusteralias'] = clustername
	clu_map['clustername'] = clustername
	if clu['quorate'] == 'true':
		clu_map['status'] = 'Quorate'
		clu_map['running'] = 'true'
	else:
		clu_map['status'] = 'Not Quorate'
		clu_map['running'] = 'false'
	clu_map['votes'] = clu['votes']
	clu_map['minquorum'] = clu['minQuorum']

	clu_map['clucfg'] = '%s?pagetype=%s&clustername=%s' \
		% (baseurl, CLUSTER_CONFIG, clustername)

	clu_map['restart_url'] = '%s?pagetype=%s&clustername=%s&task=%s' \
		% (baseurl, CLUSTER_PROCESS, clustername, CLUSTER_RESTART)
	clu_map['stop_url'] = '%s?pagetype=%s&clustername=%s&task=%s' \
		% (baseurl, CLUSTER_PROCESS, clustername, CLUSTER_STOP)
	clu_map['start_url'] = '%s?pagetype=%s&clustername=%s&task=%s' \
		% (baseurl, CLUSTER_PROCESS, clustername, CLUSTER_START)
	clu_map['delete_url'] = '%s?pagetype=%s&clustername=%s&task=%s' \
		% (baseurl, CLUSTER_PROCESS, clustername, CLUSTER_DELETE)

	svc_dict_list = list()
	for svc in svclist:
		svc_dict = {}
		svcname = svc['name']
		svc_dict['name'] = svcname
		svc_dict['servicename'] = svcname
		svc_dict['nodename'] = svc['nodename']
		svc_dict['srunning'] = svc['running']

		if svc.has_key('is_vm') and svc['is_vm'] is True:
			target_page = VM_CONFIG
		else:
			target_page = SERVICE

		svcurl = '%s?pagetype=%s&clustername=%s&servicename=%s' \
			% (baseurl, target_page, clustername, svcname)
		svc_dict['svcurl'] = svcurl
		svc_dict_list.append(svc_dict)

	clu_map['currentservices'] = svc_dict_list
	node_dict_list = list()

	for item in nodelist:
		nmap = {}
		name = item['name']
		nmap['nodename'] = name
		cfgurl = '%s?pagetype=%s&clustername=%s&nodename=%s' \
			% (baseurl, NODE, clustername, name)
		nmap['configurl'] = cfgurl
		if item['clustered'] == 'true':
			nmap['status'] = NODE_ACTIVE
		elif item['online'] == 'false':
			nmap['status'] = NODE_UNKNOWN
		else:
			nmap['status'] = NODE_INACTIVE
		node_dict_list.append(nmap)

	clu_map['currentnodes'] = node_dict_list
	return clu_map

def getNodeInfo(self, model, status, request):
	infohash = {}
	item = None
	fvars = GetReqVars(request, [ 'URL', 'clustername', 'nodename' ])
	baseurl = fvars['URL'] or LUCI_CLUSTER_BASE_URL

	nodestate = NODE_ACTIVE
	svclist = list()
	for thing in status:
		if thing['type'] == 'service':
			svclist.append(thing)

	clustername = fvars['clustername']
	nodename = fvars['nodename']
	if clustername is None or nodename is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('getNodeInfo0: %r %r' \
				% (clustername, nodename))
		return {}

	# extract correct node line from cluster status
	found = False
	for item in status:
		if (item['type'] == 'node') and (item['name'] == nodename):
			if item['online'] == 'false':
				nodestate = NODE_UNKNOWN
			elif item['clustered'] == 'true':
				nodestate = NODE_ACTIVE
			else:
				nodestate = NODE_INACTIVE
			found = True
			break

	if found is False:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('getNodeInfo1: Unable to find node "%s" in cluster status' % nodename)
		nodestate = NODE_UNKNOWN

	infohash['nodestate'] = nodestate
	infohash['nodename'] = nodename

	# set up drop down links
	if nodestate == NODE_ACTIVE:
		infohash['jl_url'] = '%s?pagetype=%s&task=%s&nodename=%s&clustername=%s' \
			% (baseurl, NODE_PROCESS, NODE_LEAVE_CLUSTER, nodename, clustername)
		infohash['reboot_url'] = '%s?pagetype=%s&task=%s&nodename=%s&clustername=%s' \
			% (baseurl, NODE_PROCESS, NODE_REBOOT, nodename, clustername)
		infohash['fence_url'] = '%s?pagetype=%s&task=%s&nodename=%s&clustername=%s' \
			% (baseurl, NODE_PROCESS, NODE_FENCE, nodename, clustername)
		infohash['delete_url'] = '%s?pagetype=%s&task=%s&nodename=%s&clustername=%s' \
			% (baseurl, NODE_PROCESS, NODE_DELETE, nodename, clustername)
	elif nodestate == NODE_INACTIVE:
		infohash['jl_url'] = '%s?pagetype=%s&task=%s&nodename=%s&clustername=%s' \
			% (baseurl, NODE_PROCESS, NODE_JOIN_CLUSTER, nodename, clustername)
		infohash['reboot_url'] = '%s?pagetype=%s&task=%s&nodename=%s&clustername=%s' \
			% (baseurl, NODE_PROCESS, NODE_REBOOT, nodename, clustername)
		infohash['fence_url'] = '%s?pagetype=%s&task=%s&nodename=%s&clustername=%s' \
			% (baseurl, NODE_PROCESS, NODE_FENCE, nodename, clustername)
		infohash['delete_url'] = '%s?pagetype=%s&task=%s&nodename=%s&clustername=%s' \
			% (baseurl, NODE_PROCESS, NODE_DELETE, nodename, clustername)
	else:
		infohash['fence_url'] = '%s?pagetype=%s&task=%s&nodename=%s&clustername=%s' \
			% (baseurl, NODE_PROCESS, NODE_FENCE, nodename, clustername)
		infohash['force_delete_url'] = '%s?pagetype=%s&task=%s&nodename=%s&clustername=%s' \
			% (baseurl, NODE_PROCESS, NODE_FORCE_DELETE, nodename, clustername)

	# figure out current services running on this node
	svc_dict_list = list()
	for svc in svclist:
		if svc['nodename'] == nodename:
			svc_dict = {}
			svcname = svc['name']
			svcurl = '%s?pagetype=%s&clustername=%s&servicename=%s' \
				% (baseurl, SERVICE, clustername, svcname)
			svc_dict['servicename'] = svcname
			svc_dict['svcurl'] = svcurl
			svc_dict_list.append(svc_dict)

	infohash['currentservices'] = svc_dict_list

	fdom_dict_list = list()
	gulm_cluster = False
	if model:
		gulm_cluster = model.getGULMPtr() is not None
		try:
			infohash['gulm_lockserver'] = model.isNodeLockserver(nodename)
		except:
			infohash['gulm_lockserver'] = False

		# next is faildoms
		fdoms = model.getFailoverDomainsForNode(nodename)
		for fdom in fdoms:
			fdom_dict = {}
			fdom_dict['name'] = fdom.getName()
			fdomurl = '%s?pagetype=%s&clustername=%s&fdomname=%s' \
				% (baseurl, FDOM, clustername, fdom.getName())
			fdom_dict['fdomurl'] = fdomurl
			fdom_dict_list.append(fdom_dict)
	else:
		infohash['gulm_lockserver'] = False

	infohash['fdoms'] = fdom_dict_list

	infohash['d_states'] = None
	nodename_resolved = resolve_nodename(self, clustername, nodename)

	if nodestate == NODE_ACTIVE or nodestate == NODE_INACTIVE:
		try:
			rc = RicciCommunicator(nodename_resolved)
			if not rc:
				raise Exception, 'connection failed'
		except Exception, e:
			rc = None
			infohash['ricci_error'] = True
			luci_log.info('Error connecting to %s: %s' \
				% (nodename_resolved, str(e)))

		if rc is not None:
			# call service module on node and find out which daemons are running
			dlist = list()
			dlist.append('ccsd')
			if not gulm_cluster:
				dlist.append('cman')
				dlist.append('fenced')
			else:
				dlist.append('lock_gulmd')
			dlist.append('rgmanager')
			states = rq.getDaemonStates(rc, dlist)
			infohash['d_states'] = states
	else:
		infohash['ricci_error'] = True

	infohash['logurl'] = '/luci/logs/?nodename=%s&clustername=%s' \
		% (nodename_resolved, clustername)
	return infohash

def getNodesInfo(self, model, status, req):
	resultlist = list()
	nodelist = list()
	svclist = list()

	fvars = GetReqVars(req, [ 'URL', 'clustername' ])
	baseurl = fvars['URL'] or LUCI_CLUSTER_BASE_URL
	clustername = fvars['clustername']

	#Sort into lists...
	for item in status:
		if item['type'] == 'node':
			nodelist.append(item)
		elif item['type'] == 'service':
			svclist.append(item)
		else:
			continue

	if clustername is None:
		try:
			clustername = model.getClusterName().strip()
		except Exception, e:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('GNI0: no cluster name: %r %s' \
					% (e, str(e)))
			return {}

	for item in nodelist:
		nl_map = {}
		name = item['name']
		nl_map['nodename'] = name

		try:
			nl_map['gulm_lockserver'] = model.isNodeLockserver(name)
		except:
			nl_map['gulm_lockserver'] = False

		cfgurl = '%s?pagetype=%s&clustername=%s&nodename=%s' \
			% (baseurl, NODE, clustername, name)
		nl_map['configurl'] = cfgurl
		nl_map['fenceurl'] = '%s#fence' % cfgurl
		if item['clustered'] == 'true':
			nl_map['status'] = NODE_ACTIVE
			nl_map['status_str'] = NODE_ACTIVE_STR
		elif item['online'] == 'false':
			nl_map['status'] = NODE_UNKNOWN
			nl_map['status_str'] = NODE_UNKNOWN_STR
		else:
			nl_map['status'] = NODE_INACTIVE
			nl_map['status_str'] = NODE_INACTIVE_STR

		nodename_resolved = resolve_nodename(self, clustername, name)

		nl_map['logurl'] = '/luci/logs?nodename=%s&clustername=%s' \
			% (nodename_resolved, clustername)

		# set up URLs for dropdown menu...
		if nl_map['status'] == NODE_ACTIVE:
			nl_map['jl_url'] = '%s?pagetype=%s&task=%s&nodename=%s&clustername=%s' \
				% (baseurl, NODE_PROCESS, NODE_LEAVE_CLUSTER, name, clustername)
			nl_map['reboot_url'] = '%s?pagetype=%s&task=%s&nodename=%s&clustername=%s' \
				% (baseurl, NODE_PROCESS, NODE_REBOOT, name, clustername)
			nl_map['fence_it_url'] = '%s?pagetype=%s&task=%s&nodename=%s&clustername=%s' \
				% (baseurl, NODE_PROCESS, NODE_FENCE, name, clustername)
			nl_map['delete_url'] = '%s?pagetype=%s&task=%s&nodename=%s&clustername=%s' \
				% (baseurl, NODE_PROCESS, NODE_DELETE, name, clustername)
		elif nl_map['status'] == NODE_INACTIVE:
			nl_map['jl_url'] = '%s?pagetype=%s&task=%s&nodename=%s&clustername=%s' \
				% (baseurl, NODE_PROCESS, NODE_JOIN_CLUSTER, name, clustername)
			nl_map['reboot_url'] = '%s?pagetype=%s&task=%s&nodename=%s&clustername=%s' \
				% (baseurl, NODE_PROCESS, NODE_REBOOT, name, clustername)
			nl_map['fence_it_url'] = '%s?pagetype=%s&task=%s&nodename=%s&clustername=%s' \
				% (baseurl, NODE_PROCESS, NODE_FENCE, name, clustername)
			nl_map['delete_url'] = '%s?pagetype=%s&task=%s&nodename=%s&clustername=%s' \
				% (baseurl, NODE_PROCESS, NODE_DELETE, name, clustername)
		else:
			nl_map['fence_it_url'] = '%s?pagetype=%s&task=%s&nodename=%s&clustername=%s' \
				% (baseurl, NODE_PROCESS, NODE_FENCE, name, clustername)
			nl_map['force_delete_url'] = '%s?pagetype=%s&task=%s&nodename=%s&clustername=%s' \
				% (baseurl, NODE_PROCESS, NODE_FORCE_DELETE, name, clustername)

		# figure out current services running on this node
		svc_dict_list = list()
		for svc in svclist:
			if svc['nodename'] == name:
				svc_dict = {}
				svcname = svc['name']
				svcurl = '%s?pagetype=%s&clustername=%s&servicename=%s' \
					% (baseurl, SERVICE, clustername, svcname)
				svc_dict['servicename'] = svcname
				svc_dict['svcurl'] = svcurl
				svc_dict_list.append(svc_dict)

		nl_map['currentservices'] = svc_dict_list

		# next is faildoms
		if model:
			fdoms = model.getFailoverDomainsForNode(name)
		else:
			nl_map['ricci_error'] = True
			fdoms = list()

		fdom_dict_list = list()
		for fdom in fdoms:
			fdom_dict = {}
			fdom_dict['name'] = fdom.getName()
			fdomurl = '%s?pagetype=%s&clustername=%s&fdomname=%s' \
				% (baseurl, FDOM, clustername, fdom.getName())
			fdom_dict['fdomurl'] = fdomurl
			fdom_dict_list.append(fdom_dict)

		nl_map['fdoms'] = fdom_dict_list
		resultlist.append(nl_map)

	return resultlist

def getFence(self, model, request):
	if not model:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('getFence0: model is None')
		return {}

	fvars = GetReqVars(request, [ 'URL', 'fencename' ])
	baseurl = fvars['URL'] or LUCI_CLUSTER_BASE_URL
	fencename = fvars['fencename']

	fence_map = {}
	nodes = model.getNodes()
	fencedevs = model.getFenceDevices()
	clustername = model.getClusterName()

	for fencedev in fencedevs:
		if fencedev.getName().strip() == fencename:
			fence_map = fencedev.getAttributes()
			try:
				fence_map['pretty_name'] = fencedev.getPrettyName()
			except:
				fence_map['unknown'] = True
				fence_map['pretty_name'] = fencedev.getAgentType()

			nodes_used = list()
			for node in nodes:
				flevels = node.getFenceLevels()
				for flevel in flevels:
					# These are the method blocks...
					kids = flevel.getChildren()
					for kid in kids:
						# These are actual devices in each level
						if kid.getName().strip() == fencedev.getName().strip():
							# See if this fd already has an entry for this node
							found_duplicate = False
							for item in nodes_used:
								if item['nodename'] == node.getName().strip():
									found_duplicate = True
							if found_duplicate is True:
								continue
							node_hash = {}
							cur_nodename = node.getName().strip()
							node_hash['nodename'] = cur_nodename
							node_hash['nodeurl'] = '%s?clustername=%s&nodename=%s&pagetype=%s' \
								% (baseurl, clustername, cur_nodename, NODE)
							nodes_used.append(node_hash)

			fence_map['nodesused'] = nodes_used
			return fence_map

	return fence_map

def getFDForInstance(fds, name):
	for fd in fds:
		if fd.getName().strip() == name:
			return fd
	raise

def getFenceInfo(self, model, request):
	if not model:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('getFenceInfo0: model is None')
		return {}

	fvars = GetReqVars(request, [ 'clustername', 'URL', 'nodename' ])
	baseurl = fvars['URL'] or LUCI_CLUSTER_BASE_URL
	clustername = fvars['clustername'] or model.getClusterName()

	nodename = fvars['nodename']
	if nodename is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('getFenceInfo1: no nodename')
		return {}

	fence_map = {}
	level1 = list() #First level fence devices
	level2 = list() #Second level fence devices
	shared1 = list() #List of available sharable fence devs not used in level1
	shared2 = list() #List of available sharable fence devs not used in level2
	fence_map['level1'] = level1
	fence_map['level2'] = level2
	fence_map['shared1'] = shared1
	fence_map['shared2'] = shared2

	major_num = 1
	minor_num = 100

	# Here we need to get fences for a node - just the first two levels
	# Each level has its own list of fence devs used in that level
	# For each fence dev, a list of instance structs is appended
	# In addition, for each level, a list of available but unused fence devs
	# is returned.
	try:
		node = model.retrieveNodeByName(nodename)
	except Exception, e:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('getFenceInfo3: unable to find node name "%s" in current node list: %r %s' % (str(nodename), e, str(e)))
		return {}

	fds = model.getFenceDevices()

	levels = node.getFenceLevels()
	len_levels = len(levels)

	if len_levels == 0:
		return fence_map

	if len_levels >= 1:
		first_level = levels[0]
		kids = first_level.getChildren()

		# This is a marker for allowing multi instances
		# beneath a fencedev
		last_kid_fd = None

		for kid in kids:
			instance_name = kid.getName().strip()
			try:
				fd = getFDForInstance(fds, instance_name)
			except:
				# Set to None in case last time thru loop
				fd = None
				continue

			if fd is not None:
				if fd.isShared() is False:
					# Not a shared dev... build struct and add
					fencedev = {}

					try:
						fencedev['prettyname'] = fd.getPrettyName()
					except:
						fencedev['unknown'] = True
						fencedev['prettyname'] = fd.getAgentType()
					fencedev['isShared'] = False
					fencedev['id'] = str(major_num)
					major_num = major_num + 1
					devattrs = fd.getAttributes()
					kees = devattrs.keys()
					for kee in kees:
						fencedev[kee] = devattrs[kee]
					kidattrs = kid.getAttributes()
					kees = kidattrs.keys()
					for kee in kees:
						if kee == 'name':
							# Don't duplicate name attr
							continue
						fencedev[kee] = kidattrs[kee]

					# This fencedev struct is complete, and needs
					# to be placed on the level1 Q. Because it is
					# non-shared, we should set last_kid_fd to none.
					last_kid_fd = None
					level1.append(fencedev)
				else:
					# This dev is shared
					if (last_kid_fd is not None) and (fd.getName().strip() == last_kid_fd['name'].strip()):
						# just append a new instance struct to last_kid_fd
						instance_struct = {}
						instance_struct['id'] = str(minor_num)
						minor_num = minor_num + 1
						kidattrs = kid.getAttributes()
						kees = kidattrs.keys()

						for kee in kees:
							if kee != 'name':
								instance_struct[kee] = kidattrs[kee]

						# Now just add this struct to last_kid_fd
						# and reset last_kid_fd
						ilist = last_kid_fd['instance_list']
						ilist.append(instance_struct)
						continue
					else:
						# Shared, but not used above...so we need
						# a new fencedev struct
						fencedev = {}
						try:
							fencedev['prettyname'] = fd.getPrettyName()
						except:
							fencedev['unknown'] = True
							fencedev['prettyname'] = fd.getAgentType()
						fencedev['isShared'] = True
						fencedev['cfgurl'] = '%s?clustername=%s&fencename=%s&pagetype=%s' \
							% (baseurl, clustername, fd.getName().strip(), FENCEDEV)
						fencedev['id'] = str(major_num)
						major_num = major_num + 1
						inlist = list()
						fencedev['instance_list'] = inlist
						devattrs = fd.getAttributes()
						kees = devattrs.keys()
						for kee in kees:
							fencedev[kee] = devattrs[kee]
						instance_struct = {}
						kidattrs = kid.getAttributes()
						kees = kidattrs.keys()
						for kee in kees:
							if kee != 'name':
								instance_struct[kee] = kidattrs[kee]

						inlist.append(instance_struct)
						level1.append(fencedev)
						last_kid_fd = fencedev
						continue
		fence_map['level1'] = level1

		# level1 list is complete now, but it is still
		# necessary to build shared1
		for fd in fds:
			isUnique = True
			if fd.isShared() is False:
				continue
			for fdev in level1:
				if fd.getName().strip() == fdev['name']:
					isUnique = False
					break
			if isUnique is True:
				shared_struct = {}
				shared_struct['name'] = fd.getName().strip()
				agentname = fd.getAgentType()
				shared_struct['agent'] = agentname
				try:
					shared_struct['prettyname'] = fd.getPrettyName()
				except:
					shared_struct['unknown'] = True
					shared_struct['prettyname'] = agentname
				shared1.append(shared_struct)
		fence_map['shared1'] = shared1

	# YUK: This next section violates the DRY rule, :-(
	if len_levels >= 2:
		second_level = levels[1]
		kids = second_level.getChildren()
		last_kid_fd = None	#This is a marker for allowing multi instances
												#beneath a fencedev
		for kid in kids:
			instance_name = kid.getName().strip()
			try:
				fd = getFDForInstance(fds, instance_name)
			except:
				fd = None #Set to None in case last time thru loop
				continue
			if fd is not None:
				if fd.isShared() is False:	#Not a shared dev...build struct and add
					fencedev = {}
					try:
						fencedev['prettyname'] = fd.getPrettyName()
					except:
						fencedev['unknown'] = True
						fencedev['prettyname'] = fd.getAgentType()
					fencedev['isShared'] = False
					fencedev['id'] = str(major_num)
					major_num = major_num + 1
					devattrs = fd.getAttributes()
					kees = devattrs.keys()
					for kee in kees:
						fencedev[kee] = devattrs[kee]
					kidattrs = kid.getAttributes()
					kees = kidattrs.keys()
					for kee in kees:
						if kee == 'name':
							continue #Don't duplicate name attr
						fencedev[kee] = kidattrs[kee]
					#This fencedev struct is complete, and needs to be placed on the
					#level2 Q. Because it is non-shared, we should set last_kid_fd
					#to none.
					last_kid_fd = None
					level2.append(fencedev)
				else:	#This dev is shared
					if (last_kid_fd is not None) and (fd.getName().strip() == last_kid_fd['name'].strip()):	#just append a new instance struct to last_kid_fd
						instance_struct = {}
						instance_struct['id'] = str(minor_num)
						minor_num = minor_num + 1
						kidattrs = kid.getAttributes()
						kees = kidattrs.keys()
						for kee in kees:
							if kee == 'name':
								continue
							instance_struct[kee] = kidattrs[kee]
						#Now just add this struct to last_kid_fd and reset last_kid_fd
						ilist = last_kid_fd['instance_list']
						ilist.append(instance_struct)
						#last_kid_fd = fd
						continue
					else: #Shared, but not used above...so we need a new fencedev struct
						fencedev = {}
						try:
							fencedev['prettyname'] = fd.getPrettyName()
						except:
							fencedev['unknown'] = True
							fencedev['prettyname'] = fd.getAgentType()
						fencedev['isShared'] = True
						fencedev['cfgurl'] = '%s?clustername=%s&fencename=%s&pagetype=%s' \
							% (baseurl, clustername, fd.getName().strip(), FENCEDEV)
						fencedev['id'] = str(major_num)
						major_num = major_num + 1
						inlist = list()
						fencedev['instance_list'] = inlist
						devattrs = fd.getAttributes()
						kees = devattrs.keys()
						for kee in kees:
							fencedev[kee] = devattrs[kee]
						instance_struct = {}
						kidattrs = kid.getAttributes()
						kees = kidattrs.keys()
						for kee in kees:
							if kee == 'name':
								continue
							instance_struct[kee] = kidattrs[kee]
						inlist.append(instance_struct)
						level2.append(fencedev)
						last_kid_fd = fencedev
						continue
		fence_map['level2'] = level2

		#level2 list is complete but like above, we need to build shared2
		for fd in fds:
			isUnique = True
			if fd.isShared() is False:
				continue
			for fdev in level2:
				if fd.getName().strip() == fdev['name']:
					isUnique = False
					break
			if isUnique is True:
				shared_struct = {}
				shared_struct['name'] = fd.getName().strip()
				agentname = fd.getAgentType()
				shared_struct['agent'] = agentname
				try:
					shared_struct['prettyname'] = fd.getPrettyName()
				except:
					shared_struct['unknown'] = True
					shared_struct['prettyname'] = agentname
				shared2.append(shared_struct)
		fence_map['shared2'] = shared2

	return fence_map

def getFencesInfo(self, model, request):
	fences_map = {}
	if not model:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('getFencesInfo0: model is None')
		fences_map['fencedevs'] = list()
		return fences_map

	fvars = GetReqVars(request, [ 'clustername', 'URL' ])
	baseurl = fvars['URL'] or LUCI_CLUSTER_BASE_URL
	clustername = fvars['clustername'] or model.getClusterName()

	fencedevs = list() #This is for the fencedev list page

	#Get list of fence devices
	fds = model.getFenceDevices()
	for fd in fds:
		#This section determines which nodes use the dev
		#create fencedev hashmap
		nodes_used = list()

		if fd.isShared() is True:
			fencedev = {}
			attr_hash = fd.getAttributes()
			kees = attr_hash.keys()

			for kee in kees:
				fencedev[kee] = attr_hash[kee] #copy attrs over
			try:
				fencedev['pretty_name'] = fd.getPrettyName()
			except:
				fencedev['unknown'] = True
				fencedev['pretty_name'] = fd.getAgentType()

			fencedev['agent'] = fd.getAgentType()
			#Add config url for this fencedev
			fencedev['cfgurl'] = '%s?clustername=%s&fencename=%s&pagetype=%s' \
				% (baseurl, clustername, fd.getName().strip(), FENCEDEV)

			nodes = model.getNodes()
			for node in nodes:
				flevels = node.getFenceLevels()
				for flevel in flevels: #These are the method blocks...
					kids = flevel.getChildren()
					for kid in kids: #These are actual devices in each level
						if kid.getName().strip() == fd.getName().strip():
							#See if this fd already has an entry for this node
							found_duplicate = False
							for item in nodes_used:
								if item['nodename'] == node.getName().strip():
									found_duplicate = True
							if found_duplicate is True:
								continue
							node_hash = {}
							cur_nodename = node.getName().strip()
							node_hash['nodename'] = cur_nodename
							node_hash['nodeurl'] = '%s?clustername=%s&nodename=%s&pagetype=%s' \
								% (baseurl, clustername, cur_nodename, NODE)
							nodes_used.append(node_hash)

			fencedev['nodesused'] = nodes_used
			fencedevs.append(fencedev)

	fences_map['fencedevs'] = fencedevs
	return fences_map

def getVMInfo(self, model, request):
	vm_map = {}
	fvars = GetReqVars(request, [ 'clustername', 'servicename', 'URL' ])

	baseurl = fvars['URL'] or LUCI_CLUSTER_BASE_URL

	clustername = fvars['clustername']
	if clustername is None:
		clustername = model.getName()

	svcname = fvars['servicename']
	if svcname is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('getVMInfo0: no service name')
		return vm_map

	try:
		cluster_os = model.getClusterOS()
		if cluster_os.find('Tikanga'):
			os_minor = int(cluster_os[cluster_os.find('5.') + 2])
			if os_minor > 1:
				vm_map['vm_migration_choice'] = True
	except:
		pass

	vm_map['formurl'] = '%s?clustername=%s&pagetype=29&servicename=%s' \
		% (baseurl, clustername, svcname)

	try:
		vm = model.retrieveVMsByName(svcname)
	except Exception, e:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug('getVMInfo1: %s: %r %s' % (svcname, e, str(e)))
		return vm_map

	attrs = vm.getAttributes()
	if attrs:
		vm_map.update(attrs)
	return vm_map

def getResourcesInfo(self, model, request):
	fvars = GetReqVars(request, [ 'clustername', 'URL' ])

	baseurl = fvars['URL'] or LUCI_CLUSTER_BASE_URL
	if fvars['clustername'] is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('getResourcesInfo missing cluster name')
		return []

	return getResources(model, baseurl)

def getClusterName(self, model):
	return model.getClusterName()

def getClusterAlias(self, model):
	if not model:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('GCA0: no model')
		return ''
	alias = model.getClusterAlias()
	if not alias:
		return model.getClusterName()
	return alias

def getModelBuilder(self, rc, isVirtualized):
	try:
		cluster_conf_node = rq.getClusterConf(rc)
		if not cluster_conf_node:
			raise Exception, 'getClusterConf returned None'
	except Exception, e:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('GMB0: unable to get cluster_conf_node in getModelBuilder: %r %s' % (e, str(e)))
		return None

	try:
		model = ModelBuilder(None, cluster_conf_node, rc.os())
		if not model:
			raise Exception, 'ModelBuilder() returned None'
		model.setIsVirtualized(isVirtualized)
	except Exception, e:
		try:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('GMB1: An error occurred while trying to get model for conf "%s": %r %s' % (cluster_conf_node.toxml(), e, str(e)))
		except:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('GMB1: ModelBuilder failed')
		return None

	return model

def getModelForCluster(self, clustername):
	from LuciDB import getRicciAgent

	rc = getRicciAgent(self, clustername)
	if not rc:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('GMFC0: no ricci agent for %s' % clustername)
		return None

	try:
		model = getModelBuilder(None, rc, rc.dom0())
		if not model:
			raise Exception, 'model is none'
		return model
	except Exception, e:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('GMFC1: unable to get model builder for %s: %r %s' % (clustername, e, str(e)))
	return None

def LuciExtractCluModel(self, request, cluster_name=None):
	model = None
	try:
		model = request.SESSION.get('model')
		model.getClusterName()
		return model
	except Exception, e:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('LECM0: %r: %r %s' \
				% (cluster_name, e, str(e)))
		model = None

	if not cluster_name:
		fvar = GetReqVars(request, [ 'clustername' ])
		cluster_name = fvar['clustername']

	if cluster_name is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('LECM1: no cluster name')
		return None

	try:
		model = getModelForCluster(self, cluster_name)
		model.getClusterName()
		return model
	except Exception, e:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('LECM2: no model for %s: %r %s' \
				% (cluster_name, e, str(e)))
	return None

def getClusterOS(self, rc):
	from HelperFunctions import resolveOSType

	clu_map = {}

	try:
		os_str = resolveOSType(rc.os())
		clu_map['os'] = os_str
		clu_map['isVirtualized'] = rc.dom0()
	except:
		# default to rhel5 if something crazy happened.
		try:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug('An error occurred while attempting to get OS/Virt info for %s -- defaulting to rhel5/False' % rc.hostname())
		except:
			# this can throw an exception if the original exception
			# is caused by rc being None or stale.
			pass
		clu_map['os'] = 'rhel5'
		clu_map['isVirtualized'] = False
	return clu_map

def getClusterConfNodes(conf_dom):
	try:
		cluster_nodes = conf_dom.getElementsByTagName('clusternode')
		return map(lambda x: str(x.getAttribute('name')), cluster_nodes)
	except Exception, e:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('GCCN0: %r %s' % (e, str(e)))
	return None
