# Copyright (C) 2006-2008 Red Hat, Inc.
#
# This program is free software; you can redistribute
# it and/or modify it under the terms of version 2 of the
# GNU General Public License as published by the
# Free Software Foundation.

from xml.dom import minidom

from ClusterModel.ClusterNode import ClusterNode

import RicciQueries as rq
from HelperFunctions import resolveOSType, send_batch_to_hosts
from LuciSyslog import get_logger
from homebase_adapters import parseHostForm
from LuciClusterActions import propagateClusterConfAsync

from LuciZopeAsync import validate_clusvc_async, get_sysinfo_async, \
	validate_auth_async

from LuciClusterInfo import getClusterInfo, \
	getModelBuilder, LuciExtractCluModel

from conga_constants import BATCH_ID, CLUNODE_CREATE_ERRORS, \
	CLUSTER_ADD, CLUSTER_CONFIG, CLUSTER_DAEMON, CLUSTER_DELETE, \
	CLUSTER_FOLDER_PATH, CLUSTER_RESTART, CLUSTER_START, CLUSTER_STOP, \
	CLUSTER_DELETE2, DISABLE_SVC_TASK, ENABLE_SVC_TASK, FDOM, FENCEDEV, \
	FENCEDEV_NODE_CONFIG, FENCEDEVS, FLAG_DESC, INSTALL_TASK, CLUSTER_PROCESS, \
	LAST_STATUS, LUCI_DEBUG_MODE, NODE, NODE_ADD, NODE_DELETE, \
	NODE_FENCE, NODE_FORCE_DELETE, NODE_JOIN_CLUSTER, NODE_LEAVE_CLUSTER, \
	NODE_REBOOT, NODES, POSSIBLE_REBOOT_MESSAGE, PRE_CFG, PRE_INSTALL, \
	PRE_JOIN, REBOOT_TASK, REDIRECT_MSG, RESOURCES, RICCI_CONNECT_FAILURE, \
	RICCI_CONNECT_FAILURE_MSG, SEND_CONF, \
	SERVICE_LIST, SERVICES, START_NODE, TASKTYPE, \
	REDIRECT_SEC, LUCI_CLUSTER_BASE_URL, FENCE_XVM_KEY_CREATE

from FenceHandler import validateNewFenceDevice, \
	validateFenceDevice, FD_VAL_SUCCESS

from ricci_communicator import RicciCommunicator, RicciError, \
	batch_status, extract_module_status

from LuciDB import manageCluster, createClusterSystems, \
	setNodeStatus, getStorageNode, getClusterFlags, del_node_flag, \
	getClusterNode, buildClusterCreateFlags, getClusterDBObj, \
	resolve_nodename, set_node_flag, getRicciAgent, \
	CLUSTER_NODE_NEED_AUTH

from LuciZope import GetReqVars

luci_log = get_logger()

def parseClusterNodes(self, request, cluster_os):
	check_certs = False
	try:
		check_certs = 'check_certs' in request.form
	except:
		check_certs = False

	download_pkgs = 1
	try:
		download_pkgs = int(request.form['download_pkgs'].strip())
	except:
		download_pkgs = 1

	clustername = None
	try:
		clustername = str(request.form['clustername'])
	except Exception, e:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('PNC00: missing cluster name: %r %s' \
				% (e, str(e)))
		clustername = None

	if clustername is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('PCN0: no cluster name was given')
		return (False, { 'errors': [ 'No cluster name was given' ]})

	shared_storage = False
	try:
		shared_storage = request.form.has_key('enable_storage')
	except:
		shared_storage = False

	reboot_nodes = request.form.has_key('reboot_nodes')

	same_node_passwds = False
	try:
		same_node_passwds = request.form.has_key('allSameCheckBox')
	except:
		same_node_passwds = False

	add_cluster = {
		'name': clustername,
		'shared_storage': shared_storage,
		'download_pkgs': download_pkgs,
		'cluster_os': cluster_os,
		'identical_passwds': same_node_passwds,
		'check_certs': check_certs,
		'reboot_nodes': reboot_nodes
	}

	system_list, incomplete, errors, messages = parseHostForm(request, check_certs)
	add_cluster['nodes'] = system_list

	for i in system_list:
		cur_system = system_list[i]

		cur_host_trusted = 'trusted' in cur_system
		cur_host = cur_system['host']

		try:
			cur_passwd = cur_system['passwd']
		except:
			cur_passwd = None

		if (cur_host_trusted or not check_certs) and cur_passwd:
			try:
				rc = RicciCommunicator(cur_host, enforce_trust=True)
				if not rc:
					raise Exception, 'connection failed'
			except Exception, e:
				cur_system['errors'] = True
				incomplete = True
				errors.append('Unable to connect to %s: %s' \
					% (cur_host, str(e)))
				if LUCI_DEBUG_MODE is True:
					luci_log.debug_verbose('PCN1: %s: %r %s' \
						% (cur_host, e, str(e)))
				continue

			prev_auth = rc.authed()
			cur_system['prev_auth'] = prev_auth

			try:
				if prev_auth:
					messages.append('Host %s is already authenticated' \
						% cur_host)
				else:
					rc.auth(cur_passwd)

				if not rc.authed():
					raise Exception, 'authentication failed'
			except Exception, e:
				cur_system['errors'] = True
				incomplete = True
				errors.append('Error authenticating to %s: %s' \
					% (cur_host, str(e)))
				if LUCI_DEBUG_MODE is True:
					luci_log.debug_verbose('PCN2: %s: %r %s' \
						% (cur_host, e, str(e)))
				continue

			cur_cluster_info = rc.cluster_info()
			if cur_cluster_info[0] or cur_cluster_info[1]:
				cur_system['errors'] = True
				incomplete = True

				if cur_cluster_info[0]:
					cur_cluster_name = cur_cluster_info[0]
				elif cur_cluster_info[1]:
					cur_cluster_name = cur_cluster_info[1]

				try:
					if not cur_system['prev_auth']:
						rc.unauth()
						del cur_system['trusted']
				except Exception, e:
					if LUCI_DEBUG_MODE is True:
						luci_log.debug_verbose('PCN3: %s: %r %s' \
							% (cur_host, e, str(e)))

				errors.append('%s reports it is a member of cluster "%s"' \
					% (cur_host, cur_cluster_name))
				if LUCI_DEBUG_MODE is True:
					luci_log.debug_verbose('PCN4: %s: already in %s cluster' \
						% (cur_host, cur_cluster_name))
				continue

			cur_host_os = resolveOSType(rc.os())
			if cluster_os is None:
				cluster_os = cur_host_os
				add_cluster['cluster_os'] = cur_host_os

			elif cluster_os != cur_host_os:
				cur_system['errors'] = True
				incomplete = True

				try:
					if not cur_system['prev_auth']:
						rc.unauth()
						del cur_system['trusted']
				except Exception, e:
					if LUCI_DEBUG_MODE is True:
						luci_log.debug_verbose('PCN5: %s: %r %s' \
							% (cur_host, e, str(e)))

				errors.append('The cluster software version on %s (%s) does not match the software on the other cluster nodes (%s)' % (cur_host, cur_host_os, cluster_os))
				if LUCI_DEBUG_MODE is True:
					luci_log.debug_verbose('PCN6: version mismatch for %s: (%s vs. %s)' \
						% (cur_host, cur_host_os, cluster_os))
				continue

	return add_cluster, incomplete, errors, messages

def validateCreateCluster(self, request):
	try:
		request.SESSION.delete('create_cluster')
	except:
		pass

	cluster_os = None
	try:
		cluster_os = request.form['cluster_os'].strip()
		if not cluster_os:
			raise Exception, 'cluster OS is blank'
	except:
		cluster_os = None

	add_cluster, incomplete, errors, messages = parseClusterNodes(self, request, cluster_os)
	clustername = add_cluster['name']

	if len(clustername) > 15:
		errors.append('A cluster\'s name must be less than 16 characters long')

	try:
		cluster_os = add_cluster['cluster_os']
	except:
		pass

	lockmanager = 'dlm'
	if cluster_os == 'rhel4':
		add_cluster['gulm_support'] = True
		if not request.form.has_key('lockmanager'):
			# The user hasn't been presented with the RHEL4
			# lock manager options yet.
			incomplete = True
		else:
			try:
				lockmanager = request.form['lockmanager'].strip()
			except:
				lockmanager = 'dlm'

	lockservers = None
	if lockmanager == 'gulm':
		add_cluster['lockmanager'] = 'gulm'
		try:
			lockservers = filter(lambda x: x.strip(), request.form['__GULM__'])
			if not lockservers or len(lockservers) < 1:
				raise Exception, 'blank'
			num_lockservers = len(lockservers)
			if not num_lockservers in (1, 3, 5):
				errors.append('You must have exactly 1, 3, or 5 GULM lock servers. You submitted %d lock servers' % num_lockservers)
		except:
			errors.append('No lock servers were given')

		if len(errors) > 0:
			try:
				ls_hash = {}
				for i in xrange(num_lockservers):
					ls_hash['server%d' % (i + 1)] = lockservers[i]
				add_cluster['gulm_lockservers'] = ls_hash
			except:
				pass

	if incomplete or len(errors) > 0:
		request.SESSION.set('create_cluster', add_cluster)
		return (False, { 'errors': errors, 'messages': messages })

	error = manageCluster(self, clustername,
				add_cluster['nodes'], add_cluster['cluster_os'])
	if error:
		errors.append('Unable to create the cluster Luci database objects')
		request.SESSION.set('create_cluster', add_cluster)
		return (False, { 'errors': errors, 'messages': messages })

	node_list = add_cluster['nodes'].keys()

	ret = send_batch_to_hosts(node_list, 10, rq.create_cluster, 
			add_cluster['cluster_os'], clustername, clustername,
			node_list, True, True, add_cluster['shared_storage'], False,
			add_cluster['download_pkgs'], lockservers,
			add_cluster['reboot_nodes'])

	batch_id_map = {}
	for i in ret.iterkeys():
		if ret[i].has_key('error'):
			msg = 'Unable to connect to the ricci agent on %s: %s' \
						% (i, ret[i]['err_msg'])
			errors.append(msg)
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose(msg)
			continue
		batch_id_map[i] = ret[i]['batch_result'][0]

	if len(batch_id_map) == 0:
		request.SESSION.set('create_cluster', add_cluster)
		return (False, { 'errors': errors, 'messages': messages })

	buildClusterCreateFlags(self, batch_id_map, clustername)
	response = request.RESPONSE
	response.redirect('%s?pagetype=%s&clustername=%s&busyfirst=true' \
		% (request['URL'], CLUSTER_CONFIG, clustername))

def validateAddClusterNode(self, request):
	import time

	try:
		request.SESSION.delete('add_node')
	except:
		pass

	check_certs = False
	try:
		check_certs = request.form.has_key('check_certs')
	except:
		check_certs = False

	download_pkgs = 1
	try:
		download_pkgs = int(request.form['download_pkgs'].strip())
	except:
		download_pkgs = 1

	reboot_nodes = request.form.has_key('reboot_nodes')

	cluster_os = None
	try:
		cluster_os = request.form['cluster_os'].strip()
		if not cluster_os:
			raise Exception, 'cluster OS is blank'
	except:
		cluster_os = None

	clustername = None
	try:
		clustername = str(request.form['clustername'])
	except:
		clustername = None

	if clustername is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VACN0: no cluster name was given')
		return (False, { 'errors': [ 'No cluster name was given' ]})

	if cluster_os is None:
		cluster_folder = None
		try:
			cluster_folder = self.restrictedTraverse('%s%s' % (CLUSTER_FOLDER_PATH, clustername))
			if not cluster_folder:
				raise Exception, 'cluster DB object is missing'
		except Exception, e:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('VACN1: %s: %r %s' \
					% (clustername, e, str(e)))
			return (False, { 'errors': [ 'The database object for %s is missing' % clustername ] })

		try:
			cluster_os = cluster_folder.manage_getProperty('cluster_os')
			if not cluster_os:
				raise Exception, 'cluster os is blank'
		except Exception, e:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('VACN2: %s: %r %s' \
					% (clustername, e, str(e)))
			cluster_os = None

		if cluster_os is None:
			try:
				cluster_ricci = getRicciAgent(self, clustername)
				cluster_os = resolveOSType(cluster_ricci.os())
			except Exception, e:
				if LUCI_DEBUG_MODE is True:
					luci_log.debug_verbose('VACN3: %s: %r %s' \
						% (clustername, e, str(e)))
				cluster_os = None

	if cluster_os is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('Unable to determine cluster OS for %s' % clustername)
		return (False, { 'errors': [ 'Unable to determine the version of the cluster suite this cluster is running' ] })

	shared_storage = False
	try:
		shared_storage = request.form.has_key('enable_storage')
	except:
		shared_storage = False

	same_node_passwds = False
	try:
		same_node_passwds = 'allSameCheckBox' in request.form
	except:
		same_node_passwds = False

	add_cluster = {
		'name': clustername,
		'shared_storage': shared_storage,
		'download_pkgs': download_pkgs,
		'cluster_os': cluster_os,
		'identical_passwds': same_node_passwds,
		'check_certs': check_certs,
		'reboot_nodes': reboot_nodes
	}

	system_list, incomplete, errors, messages = parseHostForm(request, check_certs)
	add_cluster['nodes'] = system_list
	for i in system_list:
		cur_system = system_list[i]

		cur_host_trusted = cur_system.has_key('trusted')
		cur_host = cur_system['host']

		try:
			cur_passwd = cur_system['passwd']
		except:
			cur_passwd = None

		if (cur_host_trusted or not check_certs) and cur_passwd:
			try:
				rc = RicciCommunicator(cur_host, enforce_trust=True)
				if not rc:
					raise Exception, 'connection failed'
			except Exception, e:
				cur_system['errors'] = True
				incomplete = True
				errors.append('Unable to connect to %s: %s' \
					% (cur_host, str(e)))
				if LUCI_DEBUG_MODE is True:
					luci_log.debug_verbose('VACN4: %s: %r %s' \
						% (cur_host, e, str(e)))
				continue

			prev_auth = rc.authed()
			cur_system['prev_auth'] = prev_auth

			try:
				if prev_auth:
					messages.append('Host %s is already authenticated' \
						% cur_host)
				else:
					rc.auth(cur_passwd)

				if not rc.authed():
					raise Exception, 'authentication failed'
			except Exception, e:
				cur_system['errors'] = True
				incomplete = True
				errors.append('Error authenticating to %s: %s' \
					% (cur_host, str(e)))
				if LUCI_DEBUG_MODE is True:
					luci_log.debug_verbose('VACN5: %s: %r %s' \
						% (cur_host, e, str(e)))
				continue

			cur_cluster_info = rc.cluster_info()
			if cur_cluster_info[0] or cur_cluster_info[1]:
				cur_system['errors'] = True
				incomplete = True

				if cur_cluster_info[0]:
					cur_cluster_name = cur_cluster_info[0]
				elif cur_cluster_info[1]:
					cur_cluster_name = cur_cluster_info[1]

				try:
					if not cur_system['prev_auth']:
						rc.unauth()
						del cur_system['trusted']
				except Exception, e:
					if LUCI_DEBUG_MODE is True:
						luci_log.debug_verbose('VACN6: %s: %r %s' \
							% (cur_host, e, str(e)))

				errors.append('%s reports it is already a member of cluster "%s"' % (cur_host, cur_cluster_name))
				if LUCI_DEBUG_MODE is True:
					luci_log.debug_verbose('VACN7: %s: already in %s cluster' \
						% (cur_host, cur_cluster_name))
				continue

			cur_host_os = resolveOSType(rc.os())
			if cluster_os is not None and cluster_os != cur_host_os:
				cur_system['errors'] = True
				incomplete = True

				try:
					if not cur_system['prev_auth']:
						rc.unauth()
						del cur_system['trusted']
				except Exception, e:
					if LUCI_DEBUG_MODE is True:
						luci_log.debug_verbose('VACN8: %s: %r %s' \
							% (cur_host, e, str(e)))

				errors.append('The cluster software version on %s (%s) does not match the software on the other cluster nodes (%s)' % (cur_host, cur_host_os, cluster_os))
				if LUCI_DEBUG_MODE is True:
					luci_log.debug_verbose('VACN9: version mismatch for %s: (%s vs. %s)' \
						% (cur_host, cur_host_os, cluster_os))
				continue

	if incomplete or len(errors) > 0:
		request.SESSION.set('add_node', add_cluster)
		return (False, { 'errors': errors, 'messages': messages })

	try:
		cluster_ricci = getRicciAgent(self, clustername)
		if not cluster_ricci:
			raise Exception, 'Unable to find a ricci agent for %s' % clustername
	except Exception, e:
		incomplete = True
		errors.append('Unable to contact a ricci agent for %s' % clustername)
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VACN10: %s: %r %s' \
				% (clustername, e, str(e)))

	if incomplete or len(errors) > 0:
		request.SESSION.set('add_node', add_cluster)
		return (False, { 'errors': errors, 'messages': messages })

	try:
		model = getModelBuilder(None, cluster_ricci, cluster_ricci.dom0())
		if not model:
			errors.append('Unable to retrieve the cluster configuration for %s. The configuration XML may contain errors' % clustername)
			raise Exception, 'unable to get model for %s' % clustername

		nodesptr = model.getClusterNodesPtr()
		used_ids = {}
		for i in model.getNodes():
			used_ids[int(i.getAttribute('nodeid'))] = 1
			node_name = str(i.getAttribute('name'))
			if node_name in system_list:
				system_list[node_name]['errors'] = True
				errors.append('%s is already a member of %s' \
					% (node_name, clustername))
	except Exception, e:
		incomplete = True
		errors.append('Unable to retrieve the cluster configuration for %s. The configuration XML may contain errors' % clustername)
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VACN11: %r %s' % (e, str(e)))

	if incomplete or len(errors) > 0:
		request.SESSION.set('add_node', add_cluster)
		return (False, { 'errors': errors, 'messages': messages })

	next_node_id = 1

	try:
		for x in system_list:
			i = system_list[x]

			try:
				batch_node = rq.addClusterNodeBatch(clustername,
								True,
								True,
								shared_storage,
								False,
								download_pkgs,
								model.GULM_ptr is not None,
								reboot_nodes)
				if not batch_node:
					raise Exception, 'batch is blank'
				system_list[x]['batch'] = batch_node
			except Exception, e:
				cur_system['errors'] = True
				incomplete = True

				try:
					if not cur_system['prev_auth']:
						rc.unauth()
						del cur_system['trusted']
				except Exception, e:
					if LUCI_DEBUG_MODE is True:
						luci_log.debug_verbose('VACN12: %s: %r %s' \
							% (cur_host, e, str(e)))

				errors.append('Unable to initiate cluster join for node "%s"' \
					% cur_host)
				if LUCI_DEBUG_MODE is True:
					luci_log.debug_verbose('VACN13: %s: %r %s' \
						% (cur_host, e, str(e)))
				continue

			next_node_id += 1
			new_node = ClusterNode()
			new_node.attr_hash['name'] = str(i['host'])
			new_node.attr_hash['votes'] = str(1)
			while next_node_id in used_ids:
				next_node_id += 1
			new_node.attr_hash['nodeid'] = str(next_node_id)
			nodesptr.addChild(new_node)

		if incomplete or len(errors) > 0:
			request.SESSION.set('add_node', add_cluster)
			return (False, { 'errors': errors, 'messages': messages })

		model.setModified(True)
		conf_str = str(model.exportModelAsString())
		if not conf_str:
			raise Exception, 'Unable to save the new cluster model'

		batch_number, result = rq.setClusterConf(cluster_ricci, conf_str)
		if not batch_number or not result:
			raise Exception, 'batch or result is None'
	except Exception, e:
		incomplete = True
		errors.append('Unable to generate the new cluster configuration')
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VACN14: %s' % e)

	# Propagate the new cluster.conf to the existing nodes
	# before having any of the new nodes join. If this fails,
	# abort the whole process.
	try:
		while True:
			batch_ret = cluster_ricci.batch_status(batch_number)
			code = batch_ret[0]
			if code is True:
				break
			if code == -1:
				errors.append(batch_ret[1])
				raise Exception, str(batch_ret[1])
			if code is False:
				time.sleep(0.5)
	except Exception, e:
		incomplete = True
		errors.append('Unable to update the cluster node list for %s' \
			% clustername)
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VACN15: %r %s' % (e, str(e)))

	if incomplete or len(errors) > 0:
		request.SESSION.set('add_node', add_cluster)
		return (False, { 'errors': errors, 'messages': messages })

	error = createClusterSystems(self, clustername, system_list)
	if error:
		incomplete = True
		errors.append(error)
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VACN16: %s: %s' % (clustername, error))

	if incomplete or len(errors) > 0:
		request.SESSION.set('add_node', add_cluster)
		return (False, { 'errors': errors, 'messages': messages })

	batch_id_map = {}
	for x in system_list:
		clunode = system_list[x]
		success = True

		cur_host = clunode['host']
		try:
			rc = RicciCommunicator(cur_host)
			if not rc:
				raise Exception, 'rc is None'
		except Exception, e:
			success = False
			clunode['errors'] = True
			errors.append('Unable to connect to the ricci agent on %s: %s' \
				% (cur_host, str(e)))
			if LUCI_DEBUG_MODE is True:
				luci_log.info('VACN17: Unable to connect to the ricci daemon on host %s: %r %s' % (clunode['host'], e, str(e)))

		if success:
			try:
				resultNode = rc.process_batch(clunode['batch'], async=True)
				batch_id_map[cur_host] = resultNode.getAttribute('batch_id')
			except Exception, e:
				clunode['errors'] = True
				success = False
				if LUCI_DEBUG_MODE is True:
					luci_log.debug_verbose('VACN18: %s: %r %s' \
						% (cur_host, e, str(e)))

		if not success:
			incomplete = True
			errors.append('An error occurred while attempting to add cluster node "%s"' % cur_host)

	if incomplete or len(errors) > 0:
		request.SESSION.set('add_node', add_cluster)
		return (False, { 'errors': errors, 'messages': messages })

	buildClusterCreateFlags(self, batch_id_map, clustername)
	response = request.RESPONSE
	response.redirect('%s?pagetype=%s&clustername=%s&busyfirst=true' \
		% (request['URL'], CLUSTER_CONFIG, clustername))

def validateServiceAdd(self, request):
	from LuciValidation import validate_clusvc_add

	fvar = GetReqVars(request, [ 'clustername', 'URL' ])
	baseurl = fvar['URL'] or LUCI_CLUSTER_BASE_URL

	clustername = fvar['clustername']
	if clustername is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VFE0: No cluster name')
		return (False, {'errors': ['No cluster name was given']})

	model = LuciExtractCluModel(self, request, clustername)
	if model is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('vSA1: no model')
		return (False, { 'errors': [ 'Unable to retrieve the cluster configuration for %s. The configuration XML may contain errors' % clustername ]})

	ret = validate_clusvc_add(model, request)
	if ret[0] is not True:
		return ret

	action_type = ret[1]['action_type']
	action_msg = ret[1]['action_msg']
	ret = propagateClusterConfAsync(self, model, rc=None,
			action=action_type, pmsg=action_msg)
	if ret[0] is not True:
		return ret

	response = request.RESPONSE
	response.redirect('%s?pagetype=%s&clustername=%s&busyfirst=true' \
		% (baseurl, SERVICES, clustername))

def validateResourceAdd(self, request):
	from LuciValidation import validate_clures_add

	fvar = GetReqVars(request, [ 'clustername', 'URL' ])
	baseurl = fvar['URL'] or LUCI_CLUSTER_BASE_URL

	clustername = fvar['clustername']
	if clustername is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VFE0: No cluster name')
		return (False, {'errors': ['No cluster name was given']})

	model = LuciExtractCluModel(self, request, clustername)
	if model is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('vRA1: no model')
		return (False, { 'errors': [ 'Unable to retrieve the cluster configuration for %s. The configuration XML may contain errors' % clustername ]})

	ret = validate_clures_add(model, request)
	if ret[0] is not True:
		return ret

	resname = ret[1]['res_name']
	ret = propagateClusterConfAsync(self, model, rc=None,
			action=RESOURCES, pmsg='Configuring cluster resource %s' % resname)
	if ret[0] is not True:
		return ret

	response = request.RESPONSE
	response.redirect('%s?pagetype=%s&clustername=%s&busyfirst=true' \
			% (baseurl, RESOURCES, clustername))

	return (True, { 'messages': [ 'Resource "%s" configured successfully' % resname]})

def validateConfigCluster(self, request):
	from LuciValidation import validate_config_mcast, validate_config_qdisk, \
		validate_config_fence, validate_config_gulm, validate_config_general

	configFormValidators = {
		'general': validate_config_general,
		'mcast': validate_config_mcast,
		'fence': validate_config_fence,
		'qdisk': validate_config_qdisk,
		'gulm': validate_config_gulm
	}

	errors = list()
	messages = list()
	fvar = GetReqVars(request, [ 'configtype', 'clustername', 'URL' ])
	baseurl = fvar['URL'] or LUCI_CLUSTER_BASE_URL

	clustername = fvar['clustername']
	if clustername is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VFE0: No cluster name')
		return (False, {'errors': ['No cluster name was given']})

	model = LuciExtractCluModel(self, request, clustername)
	if model is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VCC0: no model')
		return (False, { 'errors': [ 'Unable to retrieve the cluster configuration for %s. The configuration XML may contain errors' % clustername ]})

	if clustername is None:
		clustername = model.getClusterName()

	config_type = fvar['configtype']
	if config_type is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VCC1: no config type')
		return (False, {'errors': [ 'No configuration type was given' ]})

	if not configFormValidators.has_key(config_type):
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VCC2: invalid config type: %s' \
				% config_type)
		return (False, { 'errors': [ 'An invalid configuration type "%s" was submitted' % config_type ]})

	config_validator = configFormValidators[config_type]
	ret = config_validator(model, request.form)

	retcode = ret[0]
	if ret[1].has_key('errors'):
		errors.extend(ret[1]['errors'])
	if ret[1].has_key('messages'):
		messages.extend(ret[1]['messages'])

	if retcode is not True or len(errors) > 0:
		return (False, {'errors': errors, 'messages': messages})

	ret = propagateClusterConfAsync(self, model, None,
			CLUSTER_CONFIG, 'Updating cluster configuration')
	if ret[0] is not True:
		if ret[1].has_key('errors'):
			errors.extend(ret[1]['errors'])
		return (retcode, {'errors': errors, 'messages': messages})

	request.RESPONSE.redirect('%s?pagetype=%s&clustername=%s&busyfirst=true' \
		% (baseurl, CLUSTER_CONFIG, clustername))

def validateFenceAdd(self, request):
	errors = list()
	fvar = GetReqVars(request, [ 'clustername', 'URL' ])

	baseurl = fvar['URL'] or LUCI_CLUSTER_BASE_URL

	clustername = fvar['clustername']
	if clustername is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VFE0: No cluster name')
		return (False, {'errors': ['No cluster name was given']})

	model = LuciExtractCluModel(self, request, clustername)
	if model is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VFA0: no model')
		return (False, { 'errors': [ 'Unable to retrieve the cluster configuration for %s. The configuration XML may contain errors' % clustername ] })

	ret_code, ret_obj = validateNewFenceDevice(request.form, model)
	if ret_code != FD_VAL_SUCCESS:
		errors.extend(ret_obj)
		return (False, { 'errors': errors })

	ret = propagateClusterConfAsync(self, model, None,
			CLUSTER_CONFIG, 'Creating fence device "%s"' % ret_obj)
	if ret[0] is not True:
		return ret

	request.RESPONSE.redirect('%s?pagetype=%s&clustername=%s&fencename=%s&busyfirst=true' % (baseurl, FENCEDEV, clustername, ret_obj))

def validateFenceEdit(self, request):
	errors = list()

	fvar = GetReqVars(request, [ 'clustername', 'URL' ])
	baseurl = fvar['URL'] or LUCI_CLUSTER_BASE_URL

	clustername = fvar['clustername']
	if clustername is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VFE0: No cluster name')
		return (False, {'errors': ['No cluster name was given']})

	model = LuciExtractCluModel(self, request, clustername)
	if model is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VFE1: no model')
		return (False, { 'errors': [ 'Unable to retrieve the cluster configuration for %s. The configuration XML may contain errors' % clustername ] })

	# This is a fence edit situation, so the model should already have an
	# entry for this fence device.
	#
	# pass form and model to validation method, then save changes if it passes.
	error_code, retobj = validateFenceDevice(request.form, model)
	if error_code != FD_VAL_SUCCESS:
		errors.extend(retobj)
		return (False, { 'errors': errors })

	ret = propagateClusterConfAsync(self, model, None,
			CLUSTER_CONFIG, 'Updating fence device "%s"' % retobj)
	if ret[0] is not True:
		return ret

	request.RESPONSE.redirect('%s?pagetype=%s&clustername=%s&fencename=%s&busyfirst=true' % (baseurl, FENCEDEV, clustername, retobj))

def validateNodeFenceConfig(self, request):
	from LuciValidation import validate_node_fence_config

	fvar = GetReqVars(request, [ 'nodename', 'clustername', 'URL' ])
	baseurl = fvar['URL'] or LUCI_CLUSTER_BASE_URL
	nodename = fvar['nodename']

	clustername = fvar['clustername']
	if clustername is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VFE0: No cluster name')
		return (False, {'errors': ['No cluster name was given']})

	model = LuciExtractCluModel(self, request, clustername)
	if model is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('vNFC6: no model for %s' % clustername)
		return (False, { 'errors': [ 'Unable to retrieve the cluster configuration for %s. The configuration XML may contain errors' % clustername ] })

	ret = validate_node_fence_config(model, request)
	if ret[0] is not True:
		return ret

	ret = propagateClusterConfAsync(self, model, None, FENCEDEV_NODE_CONFIG,
			'Updating fence configuration for node "%s"' % nodename)
	if ret[0] is not True:
		return ret

	request.RESPONSE.redirect('%s?pagetype=%s&clustername=%s&nodename=%s&busyfirst=true' % (baseurl, NODE, clustername, nodename))

def deleteFenceDevice(self, request):
	from LuciValidation import validate_fence_del
	errors = list()

	fvar = GetReqVars(request, [ 'clustername', 'URL' ])
	baseurl = fvar['URL'] or LUCI_CLUSTER_BASE_URL

	clustername = fvar['clustername']
	if clustername is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('DFD1: no cluster name')
		return (False, {'errors': ['No cluster name was given']})

	model = LuciExtractCluModel(self, request, clustername)
	if model is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('DFD2: no model')
		return (False, { 'errors': [ 'Unable to retrieve the cluster configuration for %s. The configuration XML may contain errors' % clustername ] })

	ret = validate_fence_del(model, request)
	if ret[0] is not True:
		return ret

	ret = propagateClusterConfAsync(self, model, None, CLUSTER_CONFIG,
			'Removing fence device "%s"' % ret[1].get('name'))
	if ret[0] is not True:
		return ret

	request.RESPONSE.redirect('%s?pagetype=%s&clustername=%s&busyfirst=true' \
		% (baseurl, FENCEDEVS, clustername))

def validateDaemonProperties(self, request):
	from LuciValidation import validate_cluster_daemon_form

	fvar = GetReqVars(request,
		[ 'nodename', 'clustername', 'URL' ])
	baseurl = fvar['URL'] or LUCI_CLUSTER_BASE_URL
	clustername = fvar['clustername']
	nodename = fvar['nodename']

	ret = validate_cluster_daemon_form(self, request)
	if ret[0] is not True:
		return ret

	enable_list = ret[1]['enable_list']
	disable_list = ret[1]['disable_list']

	errors = list()

	nodename_resolved = resolve_nodename(self, clustername, nodename)
	try:
		rc = RicciCommunicator(nodename_resolved)
		if not rc:
			raise Exception, 'rc is None'
	except Exception, e:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VDP5: RC %s: %r %s' \
				% (nodename_resolved, e, str(e)))
		errors.append('Unable to connect to the ricci agent on %s to update cluster daemon properties' % nodename_resolved)
		return (False, { 'errors': errors})

	batch_id, result = rq.updateServices(rc, enable_list, disable_list)
	if batch_id is None or result is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VDP6: setCluserConf: batchid or result is None')
		errors.append('Unable to update the cluster daemon properties on node %s' % nodename_resolved)
		return (False, { 'errors': errors})

	try:
		if len(enable_list) > 0:
			status_msg = 'Updating node "%s" daemon properties: enabling "%s"' \
				% (nodename_resolved, str(enable_list)[1:-1])
		if len(disable_list) > 0:
			status_msg = 'Updating node "%s" daemon properties: disabling "%s"' \
				% (nodename_resolved, str(disable_list)[1:-1])
		set_node_flag(self, clustername, rc.hostname(), batch_id,
			CLUSTER_DAEMON, status_msg)
	except Exception, e:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VDP: %r %s' % (e, str(e)))

	if len(errors) > 0:
		return (False, {'errors': errors})

	request.RESPONSE.redirect('%s?pagetype=%s&clustername=%s&nodename=%s&busyfirst=true' % (baseurl, NODE, clustername, nodename))

def validateFdom(self, request):
	from LuciValidation import validate_fdom

	fvar = GetReqVars(request, [ 'clustername', 'name', 'oldname', 'URL' ])
	baseurl = fvar['URL'] or LUCI_CLUSTER_BASE_URL

	name = fvar['name']
	clustername = fvar['clustername']
	if clustername is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VFE0: No cluster name')
		return (False, {'errors': ['No cluster name was given']})

	model = LuciExtractCluModel(self, request, clustername)
	if model is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('validateFdom0: no model')
		return (False, { 'errors': [ 'Unable to retrieve the cluster configuration for %s. The configuration XML may contain errors' % clustername ]})

	ret = validate_fdom(model, request)
	if ret[0] is not True:
		return ret

	ret = propagateClusterConfAsync(self, model, None,
			ret[1]['action'], ret[1]['msg'])
	if ret[0] is not True:
		return ret

	request.RESPONSE.redirect('%s?pagetype=%s&clustername=%s&fdomname=%s&busyfirst=true' % (baseurl, FDOM, clustername, name))

def validateVM(self, request):
	from LuciValidation import validate_vmsvc_form

	fvar = GetReqVars(request, [ 'clustername', 'URL' ])
	baseurl = fvar['URL'] or LUCI_CLUSTER_BASE_URL

	clustername = fvar['clustername']
	if clustername is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('validateVM0: no cluster name was given')
		return (False, {'errors': [ 'No cluster name was given' ]})

	model = LuciExtractCluModel(self, request, clustername)
	if model is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('validateVM1: no model')
		return (False, { 'errors': [ 'Unable to retrieve the cluster configuration for %s. The configuration XML may contain errors' % clustername ]})

	ret = validate_vmsvc_form(model, request)
	if ret[0] is not True:
		return ret

	action_type = ret[1]['action_type']
	action_msg = ret[1]['action_msg']

	ret = propagateClusterConfAsync(self, model, None, action_type, action_msg)
	if ret[0] is not True:
		return ret

	request.RESPONSE.redirect('%s?pagetype=%s&clustername=%s&busyfirst=true' \
		% (baseurl, SERVICES, clustername))

def validate_xvm_key_dist(self, request):
	fvars = GetReqVars(request, [ '__NODE_HOSTNAME__', 'URL', 'clustername' ])

	clustername = fvars['clustername']
	if clustername is None:
		return (False, { 'errors': [ 'No cluster name was given' ]})

	host_list = fvars['__NODE_HOSTNAME__']
	if not host_list:
		return (False, { 'errors': [ 'No cluster node addresses were given' ]})

	baseurl = fvars['URL'] or LUCI_CLUSTER_BASE_URL

	try:
		import base64
		f = open('/dev/urandom', 'r')
		new_key = f.read(4096)
		f.close()
		new_key = base64.encodestring(new_key)
		if not new_key:
			raise Exception, 'base64 encode failed'
		new_key = new_key.replace('\n', '')
	except Exception, e:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VXKD0: /dev/urandom: %r %s' % (e, str(e)))
		return (False, { 'errors': [ 'Unable to create a new fence_xvm key' ]})

	errors = list()
	ret = send_batch_to_hosts(host_list, 10, rq.set_xvm_key, new_key)
	del new_key

	for i in ret.iterkeys():
		batch_num = None
		batch_res = None

		if ret[i].has_key('batch_result'):
			batch_num, batch_res = ret[i]['batch_result']

		if batch_num is None or batch_res is None:
			errors.append('fence_xvm key creation failed for node "%s"' % i)
			if LUCI_DEBUG_MODE is True:
				luci_log.info('Setting fence_xvm key for node "%s" failed' % i)
			continue

		set_node_flag(self, clustername, i, batch_num, FENCE_XVM_KEY_CREATE,
			'Creating fence_xvm key file for host "%s"' % i)

	request.RESPONSE.redirect('%s?pagetype=%s&clustername=%s&tab=2&busyfirst=true' \
		% (baseurl, CLUSTER_CONFIG, clustername))

def validate_stop_waiting(self, request):
	fvar = GetReqVars(request, [ 'clustername', 'node', 'batchid', 'refreshurl' ])

	batchid = fvar['batchid']
	node = fvar['node']
	clustername = fvar['clustername']

	if node is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VSW0: no node name given')

	if clustername is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VSW1: no cluster name was given')

	if node and clustername:
		try:
			del_node_flag(self, clustername, node, batchid)
		except Exception, e:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('VSW2: [%s %s %s] %r %s' \
					% (node, clustername, batchid, e, str(e)))
	request.RESPONSE.redirect(fvar['refreshurl'])

def process_cluster_conf_editor(self, req):
	if req.has_key('clustername'):
		clustername = req['clustername'].strip() or None
	else:
		clustername = None

	if clustername is None:
		return { 'msg': 'No cluster name was given', 'cluster_conf': '' }

	msg_list = list(('\n'))

	cc = None
	if req.has_key('new_cluster_conf'):
		cc = req['new_cluster_conf']
		msg_list.append('Checking XML validity - ')
		cc_xml = None
		try:
			cc_xml = minidom.parseString(cc)
		except:
			pass
		if cc_xml is None:
			msg_list.append('FAILED\n')
			msg_list.append('Fix the error and try again:\n')
		else:
			msg_list.append('PASSED\n')

			msg_list.append('Making sure no cluster name change has occurred - ')
			new_name = cc_xml.firstChild.getAttribute('name')
			if new_name != clustername:
				msg_list.append('FAILED\n')
				msg_list.append('Fix the error and try again:\n')
			else:
				msg_list.append('PASSED\n')
				model.setModified(True)
				msg_list.append('DONE\n')
				msg_list.append('Propagating the new cluster.conf')

				rc = getRicciAgent(self, clustername)
				if not rc:
					if LUCI_DEBUG_MODE is True:
						luci_log.debug_verbose('VFA0: unable to find a ricci agent for the %s cluster' % clustername)
					msg_list.append('\nUnable to contact a ricci agent for cluster "%s"\n\n' % clustername)
				else:
					batch_id, result = rq.setClusterConf(rc, cc_xml.toxml())
					if batch_id is None or result is None:
						if LUCI_DEBUG_MODE is True:
							luci_log.debug_verbose('VFA1: setClusterConf: batchid or result is None')
						msg_list.append('\nUnable to propagate the new cluster configuration for cluster "%s"\n\n' % clustername)
					else:
						msg_list.append(' - DONE\n')
						cc = cc_xml.toxml()
						msg_list.append('\n\nALL DONE\n\n')
	else:
		if getClusterInfo(self, None, req) == {}:
			msg_list.append('invalid cluster')
		else:
			model = LuciExtractCluModel(self, req, clustername)
			cc = model.exportModelAsString()

	return { 'msg': ''.join(msg_list), 'cluster_conf': cc }

formValidators = {
	6: validateCreateCluster,
	7: validateConfigCluster,
	15: validateAddClusterNode,
	18: validateVM,
	19: validateVM,
	21: validateServiceAdd,
	24: validateServiceAdd,
	31: validateResourceAdd,
	33: validateResourceAdd,
	41: validateFdom,
	44: validateFdom,
	51: validateFenceAdd,
	54: validateFenceEdit,
	55: validateDaemonProperties,
	57: deleteFenceDevice,
	58: validateNodeFenceConfig,
	60: validate_xvm_key_dist,
	61: validate_stop_waiting,
	80: process_cluster_conf_editor,
	1001: validate_clusvc_async,
	1002: get_sysinfo_async,
	1003: validate_auth_async,
}

def validatePost(self, request):
	try:
		pagetype = int(request.form['pagetype'])
	except Exception, e:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VP0: error: %r %s' % (e, str(e)))
		return (False, {})

	if not pagetype in formValidators:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('VP1: no handler for page type %d' % pagetype)
		return (False, {})
	else:
		return formValidators[pagetype](self, request)

def clusterTaskProcess(self, model, request):
	fvar = GetReqVars(request, [ 'task', 'clustername', 'URL' ])

	baseurl = fvar['URL'] or LUCI_CLUSTER_BASE_URL

	task = fvar['task']
	if task is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug('CTP0: no cluster task')
		return 'No cluster task was given'

	cluname = fvar['clustername']
	if not model:
		model = LuciExtractCluModel(self, request, cluname)
		if model is None:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('CTP2: No cluster model for %s' \
					% cluname)
			return 'Unable to retrieve the cluster configuration for %s. The configuration XML may contain errors.' % cluname
	elif cluname is None:
		cluname = model.getClusterName()

	redirect_page = NODES
	if task == CLUSTER_STOP:
		from LuciClusterActions import ClusterStop
		ClusterStop(self, model)
	elif task == CLUSTER_START:
		from LuciClusterActions import ClusterStart
		ClusterStart(self, model)
	elif task == CLUSTER_RESTART:
		from LuciClusterActions import ClusterStop
		ClusterStop(self, model)
		request.RESPONSE.redirect('%s?pagetype=%s&clustername=%s&task=%s&busyfirst=true' \
			% (baseurl, CLUSTER_PROCESS, cluname, CLUSTER_START))
		return
	elif task == CLUSTER_DELETE:
		from LuciClusterActions import ClusterStop
		ClusterStop(self, model)
		request.RESPONSE.redirect('%s?pagetype=%s&clustername=%s&task=%s&busyfirst=true' \
			% (baseurl, CLUSTER_PROCESS, cluname, CLUSTER_DELETE2))
		return
	elif task == CLUSTER_DELETE2:
		from LuciClusterActions import ClusterDelete
		ret = ClusterDelete(self, model)
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('CTP3: ret is %r' % ret)
		if ret is not None:
			redirect_page = ret
	else:
		return 'An unknown cluster task was requested'

	response = request.RESPONSE
	response.redirect('%s?pagetype=%s&clustername=%s&busyfirst=true' \
		% (baseurl, redirect_page, model.getClusterName()))

def nodeTaskProcess(self, model, request):
	fvar = GetReqVars(request, [ 'task', 'clustername', 'nodename', 'URL' ])

	task = fvar['task']
	nodename = fvar['nodename']
	baseurl = fvar['URL'] or LUCI_CLUSTER_BASE_URL

	clustername = fvar['clustername']
	if clustername is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug('NTP0: missing cluster name')
		return (False, { 'errors': [ 'No cluster name was given' ]})

	if nodename is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug('NTP1: missing node name')
		return (False, { 'errors': [ 'No node name was given' ]})

	if task is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug('NTP2: missing task')
		return (False, { 'errors': [ 'No node task was given' ]})

	nodename_resolved = resolve_nodename(self, clustername, nodename)
	response = request.RESPONSE

	if task != NODE_FENCE and task != NODE_FORCE_DELETE:
		# Fencing and forced deletion are the only tasks
		# for which we don't want to talk to the target node.
		try:
			rc = RicciCommunicator(nodename_resolved)
			if not rc:
				raise Exception, 'rc is None'
		except RicciError, e:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug('NTP3: ricci error from %s: %r %s' \
					% (nodename_resolved, e, str(e)))
			return (False, { 'errors': [ 'Unable to connect to the ricci agent on %s' % nodename_resolved ]})
		except Exception, e:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug('NTP4: ricci error from %s: %r %s' \
					% (nodename_resolved, e, str(e)))
			return (False, { 'errors': [ 'Unable to connect to the ricci agent on %s' % nodename_resolved ]})

		cluinfo = rc.cluster_info()
		if not cluinfo[0] and not cluinfo[1]:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug('NTP5: node %s not in a cluster (expected %s)' \
					% (nodename_resolved, clustername))
			return (False, { 'errors': [ 'Node "%s" reports it is not in a cluster' % nodename_resolved ]})

		cname = clustername.lower()
		if cname != cluinfo[0].lower() and cname != cluinfo[1].lower():
			if LUCI_DEBUG_MODE is True:
				luci_log.debug('NTP6: node %s in unknown cluster %s:%s (expected %s)' % (nodename_resolved, cluinfo[0], cluinfo[1], clustername))
			return (False, { 'errors': [ 'Node "%s" reports it in cluster "%s." We expect it to be a member of cluster "%s"' % (nodename_resolved, cluinfo[0], clustername) ]})

		if not rc.authed():
			rc = None
			try:
				snode = getStorageNode(self, nodename)
				setNodeStatus(snode, CLUSTER_NODE_NEED_AUTH)
			except:
				# we'll hit it again, and try again then
				pass

			try:
				cnode = getClusterNode(self, nodename, clustername)
				setNodeStatus(cnode, CLUSTER_NODE_NEED_AUTH)
			except:
				# we'll hit it again, and try again then
				pass

		if rc is None:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug('NTP7: node %s is not authenticated' \
					% nodename_resolved)
			return (False, { 'errors': [ 'Node "%s" is not authenticated' % nodename_resolved ]})

	if task == NODE_LEAVE_CLUSTER:
		from LuciClusterActions import NodeLeaveCluster
		if NodeLeaveCluster(self, rc, clustername, nodename_resolved) is None:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('NTP8: nodeLeave failed')
			return (False, {'errors': [ 'Node "%s" failed to leave cluster "%s"' % (nodename_resolved, clustername) ]})

		response.redirect('%s?pagetype=%s&clustername=%s&busyfirst=true' \
			% (baseurl, NODES, clustername))
	elif task == NODE_JOIN_CLUSTER:
		from LuciClusterActions import NodeJoinCluster
		if NodeJoinCluster(self, rc, clustername, nodename_resolved) is None:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('NTP9: nodeJoin failed')
			return (False, {'errors': [ 'Node "%s" failed to join cluster "%s"' % (nodename_resolved, clustername) ]})

		response.redirect('%s?pagetype=%s&clustername=%s&busyfirst=true' \
			% (baseurl, NODES, clustername))
	elif task == NODE_REBOOT:
		from LuciClusterActions import NodeReboot
		if NodeReboot(self, rc, clustername, nodename_resolved) is None:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('NTP10: nodeReboot failed')
			return (False, {'errors': [ 'Node "%s" failed to reboot' \
				% nodename_resolved ]})

		response.redirect('%s?pagetype=%s&clustername=%s&busyfirst=true' \
			% (baseurl, NODES, clustername))
	elif task == NODE_FENCE:
		from LuciClusterActions import NodeFence
		if NodeFence(self, clustername, nodename, nodename_resolved) is None:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('NTP11: nodeFencefailed')
			return (False, {'errors': [ 'Fencing of node "%s" failed' \
				% nodename_resolved]})
		response.redirect('%s?pagetype=%s&clustername=%s&busyfirst=true' \
			% (baseurl, NODES, clustername))
	elif task == NODE_DELETE:
		from LuciClusterActions import NodeDeleteFromCluster
		if NodeDeleteFromCluster(self, rc, model, clustername, nodename, nodename_resolved) is None:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('NTP12: nodeDelete failed')
			return (False, {'errors': [ 'Deletion of node "%s" from cluster "%s" failed' % (nodename_resolved, clustername) ]})

		response.redirect('%s?pagetype=%s&clustername=%s&busyfirst=true' \
			% (baseurl, NODES, clustername))
	elif task == NODE_FORCE_DELETE:
		from LuciClusterActions import NodeForceDeleteFromCluster
		if NodeForceDeleteFromCluster(self, model, clustername, nodename, nodename_resolved) is None:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('NTP13: nodeForceDelete failed')
			return (False, {'errors': [ 'Deletion of node "%s" from cluster "%s" failed' % (nodename_resolved, clustername) ]})

		response.redirect('%s?pagetype=%s&clustername=%s&busyfirst=true' \
			% (baseurl, NODES, clustername))

def getResourceInfo(model, request):
	if not model:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('GRI0: no model object in session')
		return {}

	fvars = GetReqVars(request,
				[ 'resourcename', 'type', 'ip_address', 'clustername', 'URL' ])

	baseurl = fvars['URL'] or LUCI_CLUSTER_BASE_URL

	if fvars['type'] == 'ip':
		name = fvars['ip_address']
	else:
		name = fvars['resourcename']

	if name is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('GRI1: missing res name: %s' % str(request.form.items()))
		return {}

	from LuciClusterInfo import getResourceInfo as gri
	return gri(model, name, baseurl, res=None)

def serviceRestart(self, rc, req):
	from LuciClusterActions import RestartCluSvc

	fvars = GetReqVars(req,
				[ 'clustername', 'servicename', 'nodename', 'URL' ])
	baseurl = fvars['URL'] or LUCI_CLUSTER_BASE_URL

	ret = RestartCluSvc(self, rc, fvars)
	if ret is None:
		response = req.RESPONSE
		response.redirect('%s?pagetype=%s&clustername=%s&busyfirst=true' \
			% (baseurl, SERVICE_LIST, fvars['clustername']))
	else:
		return ret

def serviceStop(self, rc, req):
	from LuciClusterActions import StopCluSvc

	fvars = GetReqVars(req,
				[ 'clustername', 'servicename', 'nodename', 'URL' ])
	baseurl = fvars['URL'] or LUCI_CLUSTER_BASE_URL

	ret = StopCluSvc(self, rc, fvars)
	if ret is None:
		response = req.RESPONSE
		response.redirect('%s?pagetype=%s&clustername=%s&busyfirst=true' \
			% (baseurl, SERVICE_LIST, fvars['clustername']))
	else:
		return ret

def serviceStart(self, rc, req):
	from LuciClusterActions import StartCluSvc

	fvars = GetReqVars(req,
				[ 'clustername', 'servicename', 'nodename', 'URL' ])
	baseurl = fvars['URL'] or LUCI_CLUSTER_BASE_URL

	ret = StartCluSvc(self, rc, fvars)
	if ret is None:
		response = req.RESPONSE
		response.redirect('%s?pagetype=%s&clustername=%s&busyfirst=true' \
			% (baseurl, SERVICE_LIST, fvars['clustername']))
	else:
		return ret

def serviceDelete(self, rc, req):
	from LuciClusterActions import DeleteCluSvc

	fvars = GetReqVars(req,
				[ 'clustername', 'servicename', 'nodename', 'URL' ])
	baseurl = fvars['URL'] or LUCI_CLUSTER_BASE_URL
	clustername = fvars['clustername']

	model = LuciExtractCluModel(self, req, clustername)
	if model is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('serviceDelete0: no model')
		return (False, { 'errors': [ 'Unable to retrieve the cluster configuration for %s. The configuration XML may contain errors' % clustername ]})

	ret = DeleteCluSvc(self, rc, fvars, model)
	if ret is None:
		response = req.RESPONSE
		response.redirect('%s?pagetype=%s&clustername=%s&busyfirst=true' \
			% (baseurl, SERVICES, clustername))
	else:
		return ret

def serviceMigrate(self, rc, req):
	from LuciClusterActions import MigrateCluSvc

	fvars = GetReqVars(req,
				[ 'clustername', 'servicename', 'nodename', 'URL' ])
	baseurl = fvars['URL'] or LUCI_CLUSTER_BASE_URL

	ret = MigrateCluSvc(self, rc, fvars)
	if ret is None:
		response = req.RESPONSE
		response.redirect('%s?pagetype=%s&clustername=%s&busyfirst=true' \
			% (baseurl, SERVICE_LIST, fvars['clustername']))
	else:
		return ret

def resourceDelete(self, rc, req):
	from LuciClusterActions import DeleteResource

	fvars = GetReqVars(req,
		[ 'clustername', 'resourcename', 'nodename', 'URL' ])
	baseurl = fvars['URL'] or LUCI_CLUSTER_BASE_URL
	clustername = fvars['clustername']

	model = LuciExtractCluModel(self, req, clustername)
	if model is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('resourceDelete0: no model')
		return (False, { 'errors': [ 'Unable to retrieve the cluster configuration for %s. The configuration XML may contain errors' % clustername ]})

	ret = DeleteResource(self, rc, model, fvars['resourcename'])
	if ret is None:
		response = req.RESPONSE
		response.redirect('%s?pagetype=%s&clustername=%s&busyfirst=true' \
			% (baseurl, RESOURCES, clustername))
	else:
		return ret

def getSystemLogs(self, req):
	from LuciClusterActions import GetSystemLogs

	fvars = GetReqVars(req, [ 'clustername', 'nodename' ])
	return GetSystemLogs(self, fvars)

def isClusterBusy(self, req):
	items = None
	busy_map = {}
	isBusy = False
	redirect_message = False
	nodereports = list()
	report_index = 0

	fvar = GetReqVars(req, [ 'clustername', 'URL' ])
	busy_map['nodereports'] = nodereports

	cluname = fvar['clustername']
	if cluname is None:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('ICB0: No cluster name, returning empty map')
		return busy_map

	items = getClusterFlags(self, cluname)
	if not items or len(items) < 1:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('ICB1: NOT BUSY: no flags for cluster %s' \
				% cluname)
		# This returns an empty map, and indicates not busy
		return busy_map

	clusterfolder = getClusterDBObj(self, cluname)

	if LUCI_DEBUG_MODE is True:
		luci_log.debug_verbose('ICB2: %s is busy: %d flags' \
			% (cluname, len(items)))
	busy_map['busy'] = 'true'

	# Ok, here is what is going on...if there is an item,
	# we need to call ricci to get a batch report.
	# This report will tell us one of three things:
	#
	# #1)	the batch task is complete...delete ManagedSystem and render
	#		normal page
	# #2)	The batch task is NOT done, so meta refresh in 5 secs and try again
	# #3)	The ricci agent has no recollection of the task,
	#		so handle like 1 above
	###
	#
	# Here is what we have to do:
	# the map should have two lists:
	#	One list of non-cluster create tasks
	#	and one of cluster create task structs
	# For each item in items, check if this is a cluster create tasktype
	# If so, call RC, and then call the batch report method
	# check for error...if error, report and then remove flag.
	# if no error, check if complete. If not complete, report status
	# If complete, report status and remove flag.

	for item in items:
		node_report = {}
		tasktype = item[1].getProperty(TASKTYPE)
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('ICB3: task type for %s = %s' \
				% (item[0], tasktype))
		
		ricci = item[0].split('____')[0]

		node_report['node'] = ricci
		node_report['tasktype'] = tasktype
		node_report['clustername'] = cluname
		node_report['report_index'] = report_index
		report_index += 1

		if tasktype == CLUSTER_ADD or tasktype == NODE_ADD:
			node_report['isnodecreation'] = True
			# Default value
			node_report['iserror'] = False
			node_report['desc'] = item[1].getProperty(FLAG_DESC)

			batch_xml = None
			# This removes the 'flag' suffix

			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('ICB4: using rc host %s for item %s' \
					% (ricci, item[0]))

			try:
				rc = RicciCommunicator(ricci)
				if not rc:
					rc = None
					if LUCI_DEBUG_MODE is True:
						luci_log.debug_verbose('ICB5: rc is none')
			except Exception, e:
				rc = None
				if LUCI_DEBUG_MODE is True:
					luci_log.debug_verbose('ICB6: RC: %s: %r %s' \
						% (cluname, e, str(e)))

			batch_id = None
			if rc is not None:
				try:
					batch_id = item[1].getProperty(BATCH_ID)
					if LUCI_DEBUG_MODE is True:
						luci_log.debug_verbose('ICB7: got batch_id %s from %s' \
							% (batch_id, item[0]))
				except Exception, e:
					if LUCI_DEBUG_MODE is True:
						luci_log.debug_verbose('ICB8: failed to get batch_id from %s: %r %s' % (item[0], e, str(e)))

				if batch_id is not None:
					node_report['batchid'] = batch_id
					try:
						batch_xml = rc.batch_report(batch_id)
						if batch_xml is not None:
							if LUCI_DEBUG_MODE is True:
								luci_log.debug_verbose('ICB9: batch_xml for %s from batch_report is not None -- getting batch status' % batch_id)
							(creation_status, total) = batch_status(batch_xml)

							try:
								if LUCI_DEBUG_MODE is True:
									luci_log.debug_verbose('ICB10: batch status returned (%d,%d)' % (creation_status, total))
							except:
								if LUCI_DEBUG_MODE is True:
									luci_log.debug_verbose('ICB11: error logging batch status return')
						else:
							if LUCI_DEBUG_MODE is True:
								luci_log.debug_verbose('ICB12: batch_xml for cluster is None')
					except Exception, e:
						if LUCI_DEBUG_MODE is True:
							luci_log.debug_verbose('ICB13: error getting batch_xml from rc.batch_report: %r %s' % (e, str(e)))
						# No contact with ricci (-1000)
						creation_status = RICCI_CONNECT_FAILURE
						# set to avoid next if statement
						batch_xml = 'bloody_failure'

			if rc is None or batch_id is None:
				if LUCI_DEBUG_MODE is True:
					luci_log.debug_verbose('ICB14: unable to connect to a ricci agent for cluster %s to get batch status')
				# No contact with ricci (-1000)
				creation_status = RICCI_CONNECT_FAILURE
				# set to avoid next if statement
				batch_xml = 'bloody_bloody_failure'

			luci_log.debug_verbose('ICB15: batch xml: %r' % batch_xml)

			if batch_xml is None:
				# The job is done and gone from queue
				if redirect_message is False:
					# We have not displayed this message yet
					node_report['desc'] = REDIRECT_MSG
					node_report['iserror'] = True
					node_report['errormessage'] = ''
					nodereports.append(node_report)
					redirect_message = True

				if LUCI_DEBUG_MODE is True:
					luci_log.debug_verbose('ICB16: batch job is done -- deleting %s' % item[0])
				clusterfolder.manage_delObjects([item[0]])
				continue

			del_db_obj = False

			if creation_status < 0:
				# an error was encountered
				if LUCI_DEBUG_MODE is True:
					luci_log.debug_verbose('ICB17: %s: CS %d for %s' \
						% (cluname, creation_status, ricci))

				if creation_status == RICCI_CONNECT_FAILURE:
					laststatus = item[1].getProperty(LAST_STATUS)
					if LUCI_DEBUG_MODE is True:
						luci_log.debug_verbose('ICB18: last statis is %d' \
							% laststatus)

					if laststatus == INSTALL_TASK:
						# The node may be rebooting
						node_report['statusindex'] = INSTALL_TASK
						node_report['statusmessage'] = '%s%s' % (RICCI_CONNECT_FAILURE_MSG, POSSIBLE_REBOOT_MESSAGE)
					elif laststatus == 0:
						# The node may be rebooting
						node_report['statusindex'] = 0
						node_report['statusmessage'] = '%s%s' % (RICCI_CONNECT_FAILURE_MSG, PRE_INSTALL)
					elif laststatus == DISABLE_SVC_TASK:
						node_report['statusindex'] = DISABLE_SVC_TASK
						node_report['statusmessage'] = '%s%s' % (RICCI_CONNECT_FAILURE_MSG, PRE_CFG)
					elif laststatus == REBOOT_TASK:
						node_report['statusindex'] = REBOOT_TASK
						node_report['statusmessage'] = '%s%s' % (RICCI_CONNECT_FAILURE_MSG, PRE_CFG)
					elif laststatus == SEND_CONF:
						node_report['statusindex'] = SEND_CONF
						node_report['statusmessage'] = '%s%s' % (RICCI_CONNECT_FAILURE_MSG, PRE_JOIN)
					elif laststatus == ENABLE_SVC_TASK:
						node_report['statusindex'] = ENABLE_SVC_TASK
						node_report['statusmessage'] = '%s%s' % (RICCI_CONNECT_FAILURE_MSG, PRE_JOIN)
					else:
						node_report['statusindex'] = 0
						node_report['statusmessage'] = '%s Install is in an unknown state' % RICCI_CONNECT_FAILURE_MSG
					nodereports.append(node_report)
					continue
				elif creation_status == -(INSTALL_TASK):
					node_report['iserror'] = True
					(err_code, err_msg) = extract_module_status(batch_xml, INSTALL_TASK)
					node_report['errormessage'] = CLUNODE_CREATE_ERRORS[INSTALL_TASK] % err_msg
					del_db_obj = True
				elif creation_status == -(DISABLE_SVC_TASK):
					node_report['iserror'] = True
					(err_code, err_msg) = extract_module_status(batch_xml, DISABLE_SVC_TASK)
					node_report['errormessage'] = CLUNODE_CREATE_ERRORS[DISABLE_SVC_TASK] % err_msg
					del_db_obj = True
				elif creation_status == -(REBOOT_TASK):
					node_report['iserror'] = True
					(err_code, err_msg) = extract_module_status(batch_xml, REBOOT_TASK)
					node_report['errormessage'] = CLUNODE_CREATE_ERRORS[REBOOT_TASK] % err_msg
					del_db_obj = True
				elif creation_status == -(SEND_CONF):
					node_report['iserror'] = True
					(err_code, err_msg) = extract_module_status(batch_xml, SEND_CONF)
					node_report['errormessage'] = CLUNODE_CREATE_ERRORS[SEND_CONF] % err_msg
				elif creation_status == -(ENABLE_SVC_TASK):
					node_report['iserror'] = True
					(err_code, err_msg) = extract_module_status(batch_xml, DISABLE_SVC_TASK)
					node_report['errormessage'] = CLUNODE_CREATE_ERRORS[ENABLE_SVC_TASK] % err_msg
				elif creation_status == -(START_NODE):
					node_report['iserror'] = True
					(err_code, err_msg) = extract_module_status(batch_xml, START_NODE)
					node_report['errormessage'] = CLUNODE_CREATE_ERRORS[START_NODE] % err_msg
				else:
					del_db_obj = True
					node_report['iserror'] = True
					node_report['errormessage'] = CLUNODE_CREATE_ERRORS[0] % ''

				try:
					if del_db_obj is True:
						if LUCI_DEBUG_MODE is True:
							luci_log.debug_verbose('ICB19: %s node creation failed for %s: %d: deleting DB entry' % (cluname, ricci, creation_status))
						clusterfolder.manage_delObjects([ricci])
						clusterfolder.manage_delObjects([item[0]])
				except Exception, e:
					if LUCI_DEBUG_MODE is True:
						luci_log.debug_verbose('ICB20: delObjects: %s: %r %s' \
							% (item[0], e, str(e)))

				nodereports.append(node_report)
				continue
			else:
				# either the batch completed successfully, or it's still running
				if creation_status == total:
					# finished...
					busy_map['busy'] = 'true'
					node_report['statusmessage'] = 'Node created successfully. %s' % REDIRECT_MSG
					node_report['statusindex'] = creation_status
					nodereports.append(node_report)

					try:
						clusterfolder.manage_delObjects([item[0]])
					except Exception, e:
						if LUCI_DEBUG_MODE is True:
							luci_log.info('ICB21: Unable to delete %s: %r %s' \
								% (item[0], e, str(e)))
					continue
				else:
					isBusy = True
					busy_map['busy'] = 'true'
					node_report['statusmessage'] = 'Node still being created'
					node_report['statusindex'] = creation_status
					nodereports.append(node_report)
					propslist = list()
					propslist.append(LAST_STATUS)

					try:
						item[1].manage_delProperties(propslist)
						item[1].manage_addProperty(LAST_STATUS, creation_status, 'int')
						if LUCI_DEBUG_MODE is True:
							luci_log.debug_verbose('ICB22: last status -> %d' \
								% creation_status)
					except Exception, e:
						if LUCI_DEBUG_MODE is True:
							luci_log.debug_verbose('ICB23: last_status err: %s %d: %r %s' % (item[0], creation_status, e, str(e)))
					continue
		else:
			node_report['isnodecreation'] = False
			# This removes the 'flag' suffix

			try:
				rc = RicciCommunicator(ricci)
			except Exception, e:
				rc = None
				finished = -1
				err_msg = ''

				if LUCI_DEBUG_MODE is True:
					luci_log.debug_verbose('ICB24: ricci error: %s: %r %s' \
						% (ricci, e, str(e)))

			if rc is not None:
				batch_num = item[1].getProperty(BATCH_ID)
				batch_res = rc.batch_status(batch_num)
				finished = batch_res[0]
				err_msg = batch_res[1]

			if finished is True or finished == -1:
				if finished == -1:
					flag_msg = err_msg
				else:
					flag_msg = ''
				flag_desc = item[1].getProperty(FLAG_DESC)
				if flag_desc is None:
					node_report['desc'] = '%s%s' % (flag_msg, REDIRECT_MSG)
				else:
					node_report['desc'] = '%s%s%s' % (flag_msg, flag_desc, REDIRECT_MSG)
				nodereports.append(node_report)

				try:
					clusterfolder.manage_delObjects([item[0]])
				except Exception, e:
					if LUCI_DEBUG_MODE is True:
						luci_log.info('ICB25: Unable to delete %s: %r %s' \
						% (item[0], e, str(e)))
				busy_map['busy'] = None
			else:
				isBusy = True
				busy_map['busy'] = 'true'
				node_report['desc'] = item[1].getProperty(FLAG_DESC)
				nodereports.append(node_report)

	if isBusy:
		part1 = req['ACTUAL_URL']
		part2 = req['QUERY_STRING']

		try:
			dex = part2.replace('&busyfirst=true', '')
			part2 = dex
		except:
			pass
		busy_map['refreshurl'] = '%d; url=%s?%s' % (REDIRECT_SEC, part1, part2)
		req['specialpagetype'] = '1'
	else:
		try:
			query = req['QUERY_STRING'].replace('&busyfirst=true', '')
			busy_map['refreshurl'] = '%d; url=%s?%s' \
				% (REDIRECT_SEC, req['ACTUAL_URL'], query)
		except:
			busy_map['refreshurl'] = '%d; url=/luci/cluster?pagetype=3' \
				% REDIRECT_SEC
	if LUCI_DEBUG_MODE is True:
		luci_log.debug_verbose('ICB26: returning busy_map: %s' % str(busy_map))
	return busy_map
